import shutil
from pathlib import Path
from typing import Optional

from pydantic import BaseModel
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm, Prompt

from services.errors import CommandExecutionException
from services.jt import render_to_file
from services.utils import get_package_dir, get_parent_folder, mkdir_p, normalize_name


class ScriptOpts(BaseModel):
    base_path: Path
    secret_key: str
    app_name: Optional[str] = None
    vite_enabled: bool = False
    users: bool = False
    tasks: bool = False
    storage: bool = False
    sql: bool = False
    html: bool = False
    web: bool = True


console = Console()


def _welcome_message():
    p = Panel.fit(
        "[bold magenta]:smile_cat: Hello and welcome to "
        " AI services [/bold magenta]",
        border_style="red",
    )
    console.print(p)


def _empty_file(filename):
    with open(filename, "w", encoding="utf-8") as f:
        pass
    return True


def ask_webapp_name(project_name: Optional[str] = None) -> str:
    if not project_name:
        parent = get_parent_folder()
        _default = normalize_name(parent)
        project_name = Prompt.ask(
            f"Write a name for this app [yellow]please, avoid spaces and capital "
            "letters[/yellow]: ",
            default=_default,
        )

    name = normalize_name(project_name)
    console.print(
        f"The final name for this app is: [bold magenta]{name}[/bold magenta]"
    )
    return name


def final_words(opts: ScriptOpts):
    p = Panel.fit(
        "[bold magenta]:smile_cat: Congrats!!!"
        f" Project [cyan]{opts.app_name}[/cyan] created[/bold magenta]",
        border_style="red",
    )
    console.print(p)

    if opts.web:
        console.print(
            " [bold magenta]To test if everything is working "
            " you can run the following command:[/]\n"
        )
        console.print("\t[bold] srv web -L[/]\n")

    if opts.vite_enabled:
        console.print(
            " To start using Vite, you can init the svelte template, into the front folder: \n"
        )
        console.print(
            "\t[bold]degit https://github.com/nuxion/services-front-svelte front[/]\n"
        )


def _alembic_files(root, app_name):
    # alembic specific
    dst = f"{root}/{app_name}"
    mkdir_p(f"{dst}/migrations/versions")
    data = {"app_name": app_name}

    render_to_file(template="alembic/env.py", dst=f"{dst}/migrations/env.py", data=data)
    render_to_file(
        template="alembic/script.py.mako", dst=f"{dst}/migrations/script.py.mako"
    )
    with open(f"{root}/alembic.ini", "a", encoding="utf-8") as f:
        f.write(
            (
                "\n\n"
                f"[{app_name}]\n"
                "sqlalchemy.url = \n"
                f"script_location = %(here)s/{app_name}/migrations/\n"
                f"models_module = {app_name}.models\n"
                f"version_table = { app_name }_version\n"
                f"db_name = default\n"
            )
        )


def _users_feature(opts: ScriptOpts):
    dst = f"{opts.base_path}/{opts.app_name}"
    data = opts.dict()
    render_to_file(template="app/users_bp.py", dst=f"{dst}/api/users.py", data=data)
    render_to_file(
        template="app/users_models.py", dst=f"{dst}/users_models.py", data=data
    )

    render_to_file(
        template="app/commands/users.py", dst=f"{dst}/commands/users.py", data=data
    )

    if opts.html:
        shutil.copy(
            f"{get_package_dir('services')}/files/login.html",
            f"{opts.base_path}/{opts.app_name}/templates/login.html",
        )


def _html_feature(opts: ScriptOpts, from_new: bool):
    dst = f"{opts.base_path}/{opts.app_name}"
    mkdir_p(f"{dst}/templates")
    data = opts.dict()
    if from_new:
        render_to_file(template="app/views.py", dst=f"{dst}/views.py", data=data)
    else:
        render_to_file(template="app/views_by_app.py", dst=f"{dst}/views.py", data=data)

    shutil.copy(
        f"{get_package_dir('services')}/files/index.html",
        f"{opts.base_path}/{opts.app_name}/templates/index.html",
    )
    if opts.vite_enabled:
        shutil.copy(
            f"{get_package_dir('services')}/files/_layout.vite.html",
            f"{opts.base_path}/{opts.app_name}/templates/_layout.html",
        )
    else:
        shutil.copy(
            f"{get_package_dir('services')}/files/_layout.default.html",
            f"{opts.base_path}/{opts.app_name}/templates/_layout.html",
        )


def create_settings(opts: ScriptOpts):
    mkdir_p(f"{opts.base_path}/server_conf")
    _empty_file(f"{opts.base_path}/server_conf/__init__.py")
    render_to_file(
        template="settings.py",
        dst=f"{opts.base_path}/server_conf/settings.py",
        data=opts.dict(),
    )


def _add_command(name, opts: ScriptOpts):
    dst = f"{opts.base_path}/{opts.app_name}"
    data = opts.dict()
    render_to_file(
        template=f"app/commands/{name}.py", dst=f"{dst}/commands/{name}.py", data=data
    )


def _sql_feature(opts: ScriptOpts):
    dst = f"{opts.base_path}/{opts.app_name}"
    data = opts.dict()

    if not Path(f"{opts.base_path}/alembic.ini").is_file():
        render_to_file(template="app/alembic.ini", dst=f"{opts.base_path}/alembic.ini")
    render_to_file(template="app/models.py", dst=f"{dst}/models.py", data=data)
    render_to_file(template="app/db.py", dst=f"{dst}/db.py", data=data)
    _alembic_files(str(opts.base_path), opts.app_name)


def create_app(opts: ScriptOpts, from_new=False):
    """
    It creates an app. Similar to django apps.

    `from_new` param is related to if is the first time that an
    app is created, then it will a use a default `views.py` file.
    Else, if it is being created after the project, it uses
    `views_by_app.py`.
    """
    dst = f"{opts.base_path}/{opts.app_name}"
    if opts.web:
        mkdir_p(f"{dst}/api")
    mkdir_p(f"{dst}/commands")

    data = opts.dict()
    # creates and generates app's package files
    _empty_file(f"{dst}/__init__.py")
    if opts.web:
        render_to_file(template="app/web.py", dst=f"{dst}/web.py", data=data)
        render_to_file(
            template="app/api_bp.py", dst=f"{dst}/api/{opts.app_name}.py", data=data
        )
    if opts.html:
        _html_feature(opts, from_new)
    if opts.sql:
        _sql_feature(opts)

    if opts.tasks and opts.web:
        render_to_file(template="app/tasks.py", dst=f"{dst}/tasks.py", data=data)

    render_to_file(template="app/managers.py", dst=f"{dst}/managers.py", data=data)
    if opts.users and opts.sql:
        _users_feature(opts)
    elif opts.users and not opts.sql:
        raise CommandExecutionException(
            "To use the users feature you have to enable the sql feature."
        )


def create_project(opts: ScriptOpts):
    """
    Entrypoint when calling `srv startproject .`
    It will creates all the neccesary for a project to start
    """

    _welcome_message()

    default_app = ask_webapp_name(opts.app_name)
    opts.app_name = default_app

    if not Path(f"{opts.base_path}/server_conf").is_dir():
        create_settings(opts)

    if opts.sql:
        render_to_file(template="app/alembic.ini", dst=f"{opts.base_path}/alembic.ini")

    create_app(opts)

    _add_command("shell", opts)

    final_words(opts)



from typing import Any, Dict

from sqlalchemy import BigInteger, Column, ForeignKey, String
from sqlalchemy.orm import declarative_mixin, declared_attr, relationship
from sqlalchemy.schema import Table

from services.users import GroupMixin, UserMixin

from .db import Base

user_group_table = Table(
    "example_user_group",
    Base.metadata,
    Column(
        "user_id",
        BigInteger,
        ForeignKey(
            "example_user.id",
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    ),
    Column(
        "group_id",
        BigInteger,
        ForeignKey(
            "example_group.id",
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    ),
)


@declarative_mixin
class HasOwnerMixin:
    # pylint: disable=no-self-argument

    @declared_attr
    def owner_id(cls):
        return Column(
            String(16),
            ForeignKey("example_user.id", ondelete="SET NULL"),
            nullable=True,
        )

    @declared_attr
    def owner(cls):
        return relationship("UserModel")


class UserModel(UserMixin, Base):
    __tablename__ = "example_user"
    __mapper_args__ = {"eager_defaults": True}

    fullname = Column(String(), nullable=True)
    groups = relationship(
        "GroupModel",
        secondary=user_group_table,
        back_populates="users",
        lazy="selectin",
    )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "username": self.username,
            "password": self.password,
            "fullname": self.fullname,
            "email": self.email,
            "is_superuser": self.is_superuser,
            "is_active": self.is_active,
            "scopes": self.scopes,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }


class GroupModel(GroupMixin, Base):
    __tablename__ = "example_group"
    users = relationship(
        "UserModel",
        secondary=user_group_table,
        back_populates="groups",
        lazy="selectin",
    )

    __mapper_args__ = {"eager_defaults": True}


from sanic import Blueprint, Request
from sanic.response import json
from sanic_ext import openapi

from services.db.web import DBHelper
from services.errors import AuthValidationFailed, WebAuthFailed
from services.security import protected
from services.security.jwtauth import JWTAuth
from services.types import JWTPayload, JWTResponse, UserLogin
from services import utils

from ..managers import UserManager

users_bp = Blueprint("users", url_prefix="users", version="v1")


@users_bp.post("/login")
@openapi.response(200, {"application/json": utils.to_schema(JWTResponse)})
@openapi.response(403, dict(msg=str), "Not Found")
@openapi.body(UserLogin)
async def login_handler(request: Request, um: UserManager, auth: JWTAuth, db: DBHelper):
    # session = db.get_session(request)
    try:
        creds = UserLogin(**request.json)
        async with db.session() as session:
            user = await um.authenticate(
                session, username=creds.username, to_verify=creds.password
            )
    except AuthValidationFailed as exc:
        raise WebAuthFailed() from exc

    # jwt = await manager.generate_token(user)
    payload = JWTPayload(custom={"usr": user.username, "scopes": user.scopes})
    tkn = await auth.generate_token(payload)
    return json(tkn.dict(), 200)


@users_bp.get("/verify")
@openapi.response(200, {"application/json": utils.to_schema(JWTResponse)})
@protected(validators=["jwt"])
async def verify_handler(request: Request):
    return json(request.ctx.token_data, 200)


@users_bp.post("/refresh_token")
@openapi.response(200, {"application/json": utils.to_schema(JWTResponse)})
@openapi.body({"application/json": utils.to_schema(JWTResponse)})
async def refresh_handler(request, auth: JWTAuth):
    if not request.app.config.JWT_ALLOW_REFRESH:
        return json(dict(msg="Not found"), 404)

    at = request.token
    rftkn = request.json.get("refresh_token")
    if not rftkn:
        raise WebAuthFailed()

    old_token = JWTResponse(access_token=at, refresh_token=rftkn)
    # redis = request.ctx.web_redis
    try:
        jwt_res = await auth.refresh_token(
            old_token.access_token, old_token.refresh_token
        )
        return json(jwt_res.dict(), 200)
    except AuthValidationFailed as e:
        raise WebAuthFailed() from e


from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from sqlalchemy import delete as sqldelete
from sqlalchemy import desc, func, select
from sqlalchemy.engine.result import ChunkedIteratorResult
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.attributes import InstrumentedAttribute

from services import types
from services.errors import DBObjectNotFound
from services.utils import get_class

ModelT = TypeVar("ModelT")


class GenericManager(Generic[ModelT]):
    model_class: str
    lookup_key: str
    active_key: Optional[str] = None

    def __init__(self):
        self._model: ModelT = get_class(self.model_class)
        self._key = self.lookup_key
        self._active_key = self.active_key

    @property
    def tablename(self) -> str:
        return self._model.__tablename__

    def _get_column(self, col_name) -> InstrumentedAttribute:
        return getattr(self._model, col_name)

    @staticmethod
    def scalars(result) -> List[ModelT]:
        return list(result.scalars())

    async def alist_actives(
        self, session, order_by=None, offset=0, limit=10
    ) -> ChunkedIteratorResult:
        stmt = select(self._model).offset(offset).limit(limit)
        attr = self._get_column(self._active_key)
        stmt = stmt.where(attr == True)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = await session.execute(stmt)
        # return list(result.scalars())
        return result

    def list_actives(
        self, session, order_by=None, offset=0, limit=10
    ) -> ChunkedIteratorResult:
        stmt = select(self._model).offset(offset).limit(limit)
        attr = self._get_column(self._active_key)
        stmt = stmt.where(attr == True)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = session.execute(stmt)
        # return list(result.scalars())
        return result

    def list(self, session, order_by=None, offset=0, limit=10) -> ChunkedIteratorResult:
        stmt = select(self._model).offset(offset).limit(limit)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = session.execute(stmt)
        # return list(result.scalars())
        return result

    async def alist(
        self, session, order_by=None, offset=0, limit=10
    ) -> ChunkedIteratorResult:
        stmt = select(self._model).offset(offset).limit(limit)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = await session.execute(stmt)
        # return list(result.scalars())
        return result

    def _obj_or_raise(self, lookup, obj: Union[ModelT, None]) -> ModelT:
        if not obj:
            raise DBObjectNotFound(self.tablename, lookup)
        return obj

    def _obj_or_raise_active(self, lookup, obj: Union[ModelT, None]) -> ModelT:
        attr = self._get_column(self._active_key)
        if not attr:
            raise DBObjectNotFound(self.tablename, lookup)

        return obj

    def _get_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = session.execute(stmt)

        return rsp.scalar_one_or_none()

    async def _aget_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = await session.execute(stmt)

        return rsp.scalar_one_or_none()

    def _commit_or_rollback(self, session) -> bool:
        try:
            session.commit()
        except IntegrityError:
            session.rollback()
            return False
        return True

    async def _acommit_or_rollback(self, session) -> bool:
        try:
            await session.commit()
        except IntegrityError:
            await session.rollback()
            return False
        return True

    def count(self, session) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        res = session.execute(stmt)
        return res.scalar()

    async def acount(self, session) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        res = await session.execute(stmt)
        return res.scalar()

    def count_actives(self, session) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        attr = self._get_column(self.active_key)
        stmt = stmt.where(attr == True)
        res = session.execute(stmt)
        return res.scalar()

    async def acount_actives(self, session) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        attr = self._get_column(self.active_key)
        stmt = stmt.where(attr == True)
        res = await session.execute(stmt)
        return res.scalar()

    def create(self, session, data: Dict[str, Any], commit=True) -> Union[ModelT, None]:
        model = self._model(**data)
        session.add(model)
        if commit:
            res = self._commit_or_rollback(session)
            if not res:
                return None
        return model

    async def acreate(
        self, session, data: Dict[str, Any], commit=True
    ) -> Union[ModelT, None]:
        model = self._model(**data)
        session.add(model)
        if commit:
            res = await self._acommit_or_rollback(session)
            if not res:
                return None
        return model

    def get_active(self, session, key: str) -> ModelT:
        _obj = self._get_one(session, key)
        return self._obj_or_raise_active(key, _obj)

    async def aget_active(self, session, key: str) -> ModelT:
        _obj = await self._aget_one(session, key)
        return self._obj_or_raise_active(key, _obj)

    def get_one(self, session, key: str) -> ModelT:
        _obj = self._get_one(session, key)
        return self._obj_or_raise(key, _obj)

    async def aget_one(self, session, key: str) -> ModelT:
        _obj = await self._aget_one(session, key)
        return self._obj_or_raise(key, _obj)

    def update(self, session, key: str, data: Dict[str, Any]):
        u = self._get_one(session, key)
        if not u:
            raise DBObjectNotFound(self.tablename, key)

        for k, v in data.items():
            setattr(u, k, v)
        u.updated_at = datetime.utcnow()
        session.add(u)

    async def aupdate(self, session, key: str, data: Dict[str, Any]):
        u = await self._aget_one(session, key)
        if not u:
            raise DBObjectNotFound(self.tablename, key)

        for k, v in data.items():
            setattr(u, k, v)
        u.updated_at = datetime.utcnow()
        session.add(u)

    def hard_delete(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        session.execute(stmt)

    async def ahard_delete(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        await session.execute(stmt)

    async def soft_delete(self, session, key: str):
        obj = self._get_one(session, key)
        setattr(obj, self._active_key, False)
        obj.updated_at = datetime.utcnow()
        session.add(obj)

    async def asoft_delete(self, session, key: str):
        obj = await self._aget_one(session, key)
        setattr(obj, self._active_key, False)
        obj.updated_at = datetime.utcnow()
        session.add(obj)



import asyncio
import sys
from getpass import getpass
from typing import Union

import click
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from services import conf, defaults, errors, types
from services.db.sqlhelper import AsyncSQL

from ..managers import GroupManager, UserManager
from ..users_models import GroupModel, UserModel

console = Console()
loop = asyncio.get_event_loop()


def get_user_manager(settings: types.Settings) -> UserManager:
    # auth = auth_from_settings(settings.SECURITY)
    gm = GroupManager()
    um = UserManager(salt=settings.SECURITY.secret_key, groups=gm)

    return um


def get_db(settings: types.Settings):
    db = AsyncSQL.from_conf(settings.DATABASES[settings.USER_DB])
    return db


async def _create_group(
    settings: types.Settings, group_name
) -> Union[GroupModel, None]:
    db = get_db(settings)
    manager = get_user_manager(settings)
    async with db.session() as session:
        group = await manager.create_group(session, group_name)
    # await db.dispose()
    return group


async def _create_user(settings: types.Settings, data, group) -> Union[UserModel, None]:
    db = get_db(settings)
    manager = get_user_manager(settings)
    try:
        async with db.session() as session:
            user = await manager.register(session, data, default_group=group)
    except errors.GroupNotFound as e:
        console.print(f"[bold red]{e}[/]")
        return None

    return user


async def _delete_user(settings: types.Settings, username, hard=False):
    db = get_db(settings)
    manager = get_user_manager(settings)
    async with db.session() as session:
        await manager.delete(session, username, hard)
        await session.commit()


async def _list_users(settings: types.Settings, is_active=True):
    db = get_db(settings)
    manager = get_user_manager(settings)
    async with db.session() as session:
        users = await manager.list(session, is_active=is_active)
    return users


@click.group(name="users")
def users_cli():
    """
    administration of users
    """
    pass


@users_cli.command(name="create")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option(
    "--is-superuser", "-S", default=False, is_flag=True, help="Is a superuser"
)
@click.option("--group", "-g", default="users", help="Group which belongs")
@click.option(
    "--create-group", default=True, is_flag=True, help="Creates group if not exist"
)
@click.option("--scopes", "-C", default="user:r:w", help="Scopes for the user")
def user_create(settings_module, is_superuser, scopes, group, create_group):
    settings = conf.load_conf(settings_module)
    name = Prompt.ask("Username")
    password = getpass("Password: ")
    repeat = getpass("Paswword (repeat): ")
    email = Prompt.ask("Email (optional)", default=None)
    if password != repeat:
        console.print("[bold red]Paswords doesn't match[/]")
        sys.exit(-1)
    if create_group:
        group_res = loop.run_until_complete(
            _create_group(
                settings,
                group,
            )
        )
        if group_res:
            console.print(f"Group: {group} created")

    result = loop.run_until_complete(
        _create_user(
            settings,
            data={
                "username": name,
                "password": password,
                "email": email,
                "scopes": scopes,
                "is_superuser": is_superuser,
            },
            group=group,
        )
    )
    if result:
        console.print(f"[bold magenta]Congrats!! user {name} created")
    else:
        sys.exit(-1)


@users_cli.command(name="delete")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option("--hard", default=False, is_flag=True, help="Hard delete of the user")
@click.argument("username")
def user_delete(settings_module, hard, username):
    settings = conf.load_conf(settings_module)
    loop.run_until_complete(_delete_user(settings, username, hard=hard))
    console.print(f"[bold magenta]User {username} deleted")


@users_cli.command(name="list")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def users_list(settings_module):
    settings = conf.load_conf(settings_module)

    result = loop.run_until_complete(_list_users(settings))
    table = Table(title="List of users")
    table.add_column("id")
    table.add_column("username")
    table.add_column("scopes")
    table.add_column("email")
    table.add_column("is_superuser")
    table.add_column("created at")

    for user in result:
        table.add_row(
            f"{user.id}",
            user.username,
            user.scopes,
            f"{user.email}",
            f"{user.is_superuser}",
            f"{user.created_at}",
        )

    console.print(table)


import datetime

from sqlalchemy import DateTime, BigInteger
from sqlalchemy.orm import Mapped, declarative_mixin, mapped_column, MappedAsDataclass
from sqlalchemy.sql import functions

# from sqlalchemy.types import BINARY, DECIMAL, JSON, BigInteger


@declarative_mixin
class DateTrackMixin:
    created_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )


@declarative_mixin
class DateTrackMixinAsDataclass(MappedAsDataclass):
    created_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(), server_default=functions.now(), nullable=False, init=False
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(), server_default=functions.now(), nullable=False, init=False
    )


@declarative_mixin
class BaseMixinAsDataclass(MappedAsDataclass):
    id: Mapped[int] = mapped_column(BigInteger, primary_key=True, init=False)
    created_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(),
        index=True,
        init=False,
        server_default=functions.now(),
        nullable=False,
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(),
        init=False,
        server_default=functions.now(),
        nullable=False,
    )



import sys
from services import defaults, types

# WARNING:
# We do our best effort to keep sensible information private
# but in the scenario of an intrusion into the network or machines
# where servers are hosted, then it represents a risk for the information
# stored in those machines.

# General Folders for the server
DATABASES = {
    "default": types.Database(
        name="default",
        async_url="sqlite+aiosqlite:///db.sqlite",
        sync_url="sqlite:///db.sqlite",
        description="Default database",
    )
}

APPS = [
    "{{ data.app_name }}.web.WebApp",
]

SECURITY = types.SecurityConfig(
    secret_key="{{ data.secret_key }}",
    jwt=types.JWTConfig(
        alg="HS256",
        secret="{{ data.secret_key }}"
    )
)

{% if data.vite_enabled -%}
VITE_ENABLED = True
VITE_CONFIG = types.ViteConfig(
    # Where dynamic assets in vite are configurated
    VITE_STATIC_URL_PATH="assets",
    # Local path from the python app where the vite dynamic assets are
    VITE_STATIC_DIR="front/src/assets",
    # where vite put the buil
    VITE_OUTPUT_DIR="front/dist",
    # where vite dev server is listening
    VITE_DEV_SERVER="http://localhost:5173",
    # if dev mode is true, the it will render the hmr script tag
    VITE_DEV_MODE=True,
    # special case for react
    VITE_REACT_MODE=False,
    # base url in the vite server the same as
    #    https://vitejs.dev/config/server-options.html#server-base
    VITE_BASE="/"
)

STATICFILES_DIRS = [
        types.StaticDir(
            name="public",
            uripath="",
            localdir="front/public/"
        )
]
{% endif -%}

{% if data.tasks -%}
TASKS = types.TasksBackend(
    uri="sqlite+aiosqlite:///tasks.sqlite",
    backend_class="services.ext.sql.workers.SQLBackend"

)
{% endif -%}

{% if data.storage -%}
STORAGE = {
    "default": types.Storage(
        bucket=".storage", store_class="services.storage.AsyncLocal"
    )
}
{% endif -%}


TEMPLATES_DIR = [
    "{{ data.app_name }}/templates"
]


COMMANDS = [
    "{{ data.app_name }}.commands.shell.shell_cli",
    {% if data.web -%}
    "services.commands.web_cli",
    "services.commands.create_app_cli",
    {% if data.tasks -%}
    "services.commands.tasks.tasks_cli",
    {% endif -%}
    {% endif -%}
    {% if data.sql -%}
    "services.commands.db_cli",
    {% endif -%}
    {% if data.users -%}
    "{{ data.app_name }}.commands.users.users_cli",
    {% endif -%}

]



from sanic import Blueprint, Request
from sanic.response import html, redirect
from sanic.views import HTTPMethodView
from sanic_ext import openapi

{% if data.users -%}
from services.db.web import DBHelper
from services.errors import AuthValidationFailed
from services.errors.web import WebAuthFailed
from services.security.sessionauth import SessionAuth
{% endif -%}
from services.shortcuts import async_render
from services.types import HtmlData, UserLogin

from .managers import UserManager

web_bp = Blueprint("web")

def new_data(content):
    return HtmlData(ctx={"DEV": True}, title="{{ data.app_name }}", content=content)

{% if data.users -%}
@web_bp.get("/")
@openapi.exclude()
async def default_handler(request: Request, auth: SessionAuth):
    user = auth.get_username(request)
    data = new_data(dict(msg="hello", username=user))
    text = await async_render(request, "index.html", data)
    return html(text)


@web_bp.get("/logout")
@openapi.exclude()
async def logout_handler(request: Request, auth: SessionAuth):
    response = redirect("/")
    user = auth.get_username(request)
    if user:
        response = auth.unset_cookie(response)
    return response


class LoginView(HTTPMethodView):
    async def get(self, request):
        data = new_data(dict(msg="hello"))
        text = await async_render(request, "login.html", data)
        return html(text)

    async def post(self, request, auth: SessionAuth, um: UserManager, db: DBHelper):
        try:
            creds = UserLogin(
                username=request.form.get("username"),
                password=request.form.get("password"),
            )
            async with db.session() as session:
                user = await um.authenticate(
                    session, username=creds.username, to_verify=creds.password
                )
        except AuthValidationFailed as exc:
            raise WebAuthFailed() from exc

        response = redirect("/")
        auth.set_cookie(response, user.username)
        return response


class RegisterView(HTTPMethodView):
    async def get(self, request):
        data = new_data(dict(msg="hello"))
        text = await async_render(request, "login.html", data)
        return html(text)

    async def post(self, request, auth: SessionAuth, um: UserManager, db: DBHelper):
        try:
            creds = UserLogin(
                username=request.form.get("username"),
                password=request.form.get("password"),
            )
            async with db.session() as session:
                user = await um.authenticate(
                    session, username=creds.username, to_verify=creds.password
                )
        except AuthValidationFailed as exc:
            raise WebAuthFailed() from exc

        response = redirect("/")
        auth.set_cookie(response, user.username)
        return response


web_bp.add_route(LoginView.as_view(), "/login")
{% else %}
@web_bp.get("/")
@openapi.exclude()
async def default_handler(request: Request):
    return text("Hello world")

{% endif -%}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ai-services'
copyright = '2023, Xavier Petit'
author = 'Xavier Petit'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.autosectionlabel", "myst_parser"]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__'
}



from typing import List, Union

from itsdangerous import TimestampSigner
from itsdangerous.exc import BadSignature

from services import types
from services.security import IAuth


class SessionAuth(IAuth):
    def __init__(self, conf: types.SecurityConfig, cookie_name="SSID", secure=True):
        self.conf = conf
        # self.store = store
        self.cookie = cookie_name
        self.sign = TimestampSigner(conf.secret_key, salt=cookie_name)
        self._secure = secure

    def _get_cookie(self, request) -> Union[str, None]:
        cookie = request.cookies.get(self.cookie)
        return cookie

    def get_username(self, request) -> Union[str, None]:
        username = self.decode_request(request)
        return username

    def encode(self, username):
        svalue = self.sign.sign(username).decode("utf-8")
        return svalue

    def set_cookie(self, response, username):
        svalue = self.sign.sign(username).decode("utf-8")
        response.cookies[self.cookie] = svalue
        response.cookies[self.cookie]["domain"] = self.conf.domain
        response.cookies[self.cookie]["httponly"] = True
        response.cookies[self.cookie]["secure"] = self._secure
        response.cookies[self.cookie]["samesite"] = "lax"
        response.cookies[self.cookie]["max-age"] = self.conf.ttl_refresh_token
        return response

    def unset_cookie(self, response):
        response.cookies[self.cookie] = "expired"
        response.cookies[self.cookie]["max-age"] = 0
        return response

    def decode_request(self, request) -> Union[str, None]:
        final = None
        value = self._get_cookie(request)
        if not value:
            return None
        try:
            _final = self.sign.unsign(value, max_age=self.conf.ttl_refresh_token)
            final = _final.decode("utf-8")
        except BadSignature:
            pass
        return final

    def validate_request(self, request, policies: List[str] = None, require_all=True):
        decoded = self.decode_request(request)
        if decoded:
            request.ctx.username = decoded
            request.ctx.is_authenticated = True
            request.ctx.is_authorized = True

    def update_response(self, request, response):
        decoded = self.decode_request(request)
        res = self.set_cookie(response, decoded)
        return res



import contextlib
import logging
from typing import Any, Callable, Coroutine, List, Optional

from sqlalchemy import Table, create_engine, event, inspect, text
from sqlalchemy.engine.base import Engine
from sqlalchemy.exc import IntegrityError, NoSuchTableError
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine

# from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import (
    DropConstraint,
    DropTable,
    ForeignKeyConstraint,
    MetaData,
    Table,
)

from services import types


def drop_everything(engine):
    """(On a live db) drops all foreign key constraints before dropping all tables.
    Workaround for SQLAlchemy not doing DROP ## CASCADE for drop_all()
    (https://github.com/pallets/flask-sqlalchemy/issues/722)
    """

    con = engine.connect()
    trans = con.begin()
    inspector = inspect(engine)

    # We need to re-create a minimal metadata with only the required things to
    # successfully emit drop constraints and tables commands for postgres (based
    # on the actual schema of the running instance)
    meta = MetaData()
    tables = []
    all_fkeys = []

    for table_name in inspector.get_table_names():
        fkeys = []

        for fkey in inspector.get_foreign_keys(table_name):
            if not fkey["name"]:
                continue

            fkeys.append(ForeignKeyConstraint((), (), name=fkey["name"]))

        tables.append(Table(table_name, meta, *fkeys))
        all_fkeys.extend(fkeys)

    for fkey in all_fkeys:
        con.execute(DropConstraint(fkey))

    for table in tables:
        con.execute(DropTable(table))

    trans.commit()


def sqlite_async_uri(filename) -> str:
    return f"sqlite+aiosqlite:///{filename}"


async def async_vacuum(engine: AsyncEngine, table: Optional[str] = None) -> bool:
    """https://github.com/sqlalchemy/sqlalchemy/discussions/6959"""
    autocommit = engine.execution_options(isolation_level="AUTOCOMMIT")
    _vacuum = None
    url = str(engine.url)
    if "postgresql" in url and table:
        _vacuum = f"VACUUM ANALYZE {table}"
    elif "sqlite" in url:
        _vacuum = "vacuum"

    if _vacuum:
        async with autocommit.connect() as conn:
            await conn.execute(text(_vacuum))
        await autocommit.dispose()
        return True
    return False


async def async_set_pragma(engine: AsyncEngine, pragma="journal_mode=WAL") -> bool:
    """https://github.com/sqlalchemy/sqlalchemy/discussions/6959"""
    autocommit = engine.execution_options(isolation_level="AUTOCOMMIT")
    _pragma = None
    url = str(engine.url)
    if "sqlite" in url:
        _pragma = f"PRAGMA {pragma}"
    if _pragma:
        async with autocommit.connect() as conn:
            await conn.execute(text(_pragma))
        await autocommit.dispose()
        return True
    return False


def set_pragma(engine: Engine, pragma="journal_mode=WAL") -> bool:
    """https://github.com/sqlalchemy/sqlalchemy/discussions/6959"""
    autocommit = engine.execution_options(isolation_level="AUTOCOMMIT")
    _pragma = None
    url = str(engine.url)
    if "sqlite" in url:
        _pragma = f"PRAGMA {pragma}"

    if _pragma:
        with autocommit.connect() as conn:
            conn.execute(text(_pragma))
        return True
    return False


def vacuum(engine: Engine, table: Optional[str] = None) -> bool:
    """https://github.com/sqlalchemy/sqlalchemy/discussions/6959"""
    autocommit = engine.execution_options(isolation_level="AUTOCOMMIT")
    _vacuum = None
    url = str(engine.url)
    if "postgresql" in url and table:
        _vacuum = f"VACUUM ANALYZE {table}"
    elif "sqlite" in url:
        _vacuum = "vacuum"

    if _vacuum:
        with autocommit.connect() as conn:
            conn.execute(text(_vacuum))
        return True
    return False


class SQL:
    def __init__(
        self, db: types.Database, engine: Engine, autoflush=True, expire_on_commit=False
    ):
        """SQL sync helper to build connections and sessions
        it's prepared to be used with Core and ORM

        :param db: accepts type :class:`services.types.Database`
        :type db: services.types.Database
        """
        self.conf = db
        self._autoflush = autoflush
        self._engine = engine
        self._session = sessionmaker(self._engine, expire_on_commit=expire_on_commit)
        self._tables: List[str] = None

    @staticmethod
    def create_engine(conf: types.Database) -> Engine:
        if "sqlite" in conf.sync_url.split("://", maxsplit=1)[0]:
            engine = create_engine(conf.sync_url, echo=conf.debug)
        else:
            engine = create_engine(
                conf.sync_url,
                pool_size=conf.pool_size,
                max_overflow=conf.max_overflow,
                echo=conf.debug,
            )
        return engine

    @property
    def engine(self) -> Engine:
        return self._engine

    @property
    def session(self) -> sessionmaker:
        """
        It's a session factory it should be used as:

        .. code:

            with db.session() as session:
                session.execute(...)

        """
        return self._session

    @classmethod
    def from_uri(cls, uri: str, autoflush=True, expire_on_commit=False) -> "SQL":
        _db = types.Database(async_url=uri)
        engine = cls.create_engine(_db)
        obj = cls(
            _db, engine=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
        )
        # await obj.init()
        return obj

    @classmethod
    def from_conf(
        cls, db: types.Database, autoflush=True, expire_on_commit=False
    ) -> "SQL":
        engine = cls.create_engine(db)
        obj = cls(
            db, engine=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
        )
        return obj

    def create_all(self, meta: MetaData):
        """
        To create a database it's need the MetaData object
        from sqlalchemy
        """
        meta.create_all(self.engine)

    def drop_all(self, meta: Optional[MetaData] = None, all_=True):
        """it drops the content of a database, if all_ is set False
        then it only will drop elementes in meta.

        """

        if all_:
            drop_everything(self.engine)
        elif meta:
            meta.drop_all(self.engine)
        else:
            raise AttributeError("MetaData object or all_ param should be provided")

    def tables(self) -> List[str]:
        if not self._tables:
            self._tables = self.list_tables()
        return self._tables

    def list_tables(self) -> List[str]:
        inspector = inspect(self.engine)
        return inspector.get_table_names()

    def reflect_metadata(self) -> MetaData:
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        return metadata

    def get_table(self, table_name: str, meta=None) -> Table:
        if not meta:
            meta = MetaData()
        # try:
        table = Table(table_name, meta, autoload_with=self.engine)
        # except NoSuchTableError:
        return table

    def add_event(self, func: Callable, type_event="connect"):
        event.listen(self.engine, type_event, func)

    def session_factory(self, expire_on_commit=False) -> sessionmaker:
        """it's actually returns a session"""
        return sessionmaker(self.engine, expire_on_commit=expire_on_commit)


def delete_table(db: SQL, table_name: str) -> bool:
    try:
        tbl = db.get_table(table_name)
        tbl.drop(bind=db.engine)
    except NoSuchTableError:
        return False
    return True


class AsyncSQL:
    """
    As example checks:
    https://docs.sqlalchemy.org/en/14/_modules/examples/asyncio/async_orm.html

    scoped_session, useful in thread environments is removed in Async,
    AsyncSession is prefered insted: https://docs.sqlalchemy.org/en/14/orm/extensions/asyncio.html#using-asyncio-scoped-session
    """

    # Meta = MetaData()

    def __init__(
        self,
        db: types.Database,
        engine: AsyncEngine,
        autoflush=True,
        expire_on_commit=False,
    ):
        self.conf = db
        self._autoflush = autoflush
        self._engine = engine
        self._async_session = async_sessionmaker(
            self._engine, expire_on_commit=expire_on_commit
        )
        self._tables: List[str] = None

    @property
    def async_session(self) -> async_sessionmaker[AsyncSession]:
        """
        It's a session factory it should be used as:

        .. code:

            async with db.async_session() as session:
                await session.execute(...)

        """
        return self._async_session

    @staticmethod
    def create_engine(conf: types.Database) -> AsyncEngine:
        if "sqlite" in conf.async_url.split("://", maxsplit=1)[0]:
            engine = create_async_engine(conf.async_url, echo=conf.debug)
        else:
            engine = create_async_engine(
                conf.async_url,
                pool_size=conf.pool_size,
                max_overflow=conf.max_overflow,
                echo=conf.debug,
            )

        return engine

    @classmethod
    def from_uri(cls, uri: str, autoflush=True, expire_on_commit=False) -> "AsyncSQL":
        _db = types.Database(async_url=uri)
        engine = cls.create_engine(_db)
        obj = cls(
            _db, engine=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
        )
        # await obj.init()
        return obj

    @classmethod
    def from_conf(
        cls, db: types.Database, autoflush=True, expire_on_commit=False
    ) -> "AsyncSQL":
        engine = cls.create_engine(db)
        obj = cls(
            db, engine=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
        )
        return obj

    @contextlib.asynccontextmanager
    async def conn(self):
        try:
            async with self._engine.connect() as conn:
                yield conn
        finally:
            await self._engine.dispose()

    @contextlib.asynccontextmanager
    async def begin(self):
        try:
            async with self._engine.begin() as conn:
                yield conn
        finally:
            await self._engine.dispose()

    @contextlib.asynccontextmanager
    async def session(self):
        try:
            async with self.async_session() as session:
                yield session
        finally:
            await self._engine.dispose()

    def session_factory(
        self, autoflush=True, expire_on_commit=False
    ) -> async_sessionmaker:
        """
        expire_on_commit=False will prevent attributes from being expired
        after commit
        """
        return async_sessionmaker(
            self._engine,
            autoflush=autoflush,
            expire_on_commit=expire_on_commit,
        )

    def add_event(self, func: Callable, type_event="connect"):
        event.listen(self.engine.sync_engine, type_event, func)

    @property
    def engine(self) -> AsyncEngine:
        return self._engine

    @property
    def sync_engine(self) -> Engine:
        return self._engine.sync_engine

    async def create_all(self, meta: MetaData):
        async with self._engine.begin() as conn:
            await conn.run_sync(meta.create_all)

    async def drop_all(self, meta: Optional[MetaData] = None, all_=True):
        async with self._engine.begin() as conn:
            if all_:
                await conn.run_sync(drop_everything, self.sync_engine)
            elif meta:
                await conn.run_sync(meta.drop_all)
            else:
                raise AttributeError("MetaData object or all_ param should be provided")

    async def tables(self) -> Coroutine[Any, Any, List[str]]:
        if not self._tables:
            self._tables = await self.list_tables()
        return self._tables

    async def list_tables(self) -> Coroutine[Any, Any, List[str]]:
        async with self._engine.connect() as conn:
            tables = await conn.run_sync(
                lambda sync_conn: inspect(sync_conn).get_table_names()
            )
        return tables

    async def dispose(self):
        # for AsyncEngine created in function scope, close and
        # clean-up pooled connections
        await self._engine.dispose()


def commit_or_rollback(session) -> bool:
    try:
        session.commit()
    except IntegrityError:
        session.rollback()
        return False
    return True


async def acommit_or_rollback(session) -> bool:
    try:
        await session.commit()
    except IntegrityError:
        session.rollback()
        return False
    return True



import importlib
import logging
import logging.config
import os
import subprocess
import sys

from services import defaults
from services.types import Settings

GLOBAL_MODULE = "services.global_settings"
ENVIRONMENT_VARIABLE = "SRV_SETTINGS_MODULE"
DEFAULT_MODULE = os.environ.get(ENVIRONMENT_VARIABLE, GLOBAL_MODULE)


def _get_level(level):
    return getattr(logging, level)


def _execute_cmd(cmd) -> str:
    """Wrapper around subprocess"""
    with subprocess.Popen(
        cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:
        out, err = p.communicate()
        if err:
            raise AttributeError(err.decode())

        return out.decode().strip()


def define_base_path() -> str:
    """
    It tries to determine a base path for BASE_PATH in settings.
    If a environment var exist for this, then it will use that, else
    it will use git, if fail it will go to an upper level.
    """
    root_dir = os.getcwd()

    base_path = os.environ.get(defaults.BASE_PATH_ENV, root_dir)

    return base_path


def load_conf(settings_module=None, default_module=GLOBAL_MODULE) -> Settings:
    """
    It loads the settings modules in the following order of priority:

    1. settings_module function param
    2. environment variable of defaults.SETTINGS_MODULE_VAR
    3. DEFAULT_MODULE value.

    """
    if not settings_module:
        settings_module = os.getenv(defaults.SETTINGS_MODULE_VAR, default_module)
    module_loaded = settings_module
    base_path = define_base_path()
    sys.path.insert(0, base_path)

    try:
        mod = importlib.import_module(settings_module)
    except ModuleNotFoundError:
        mod = importlib.import_module(GLOBAL_MODULE)
        module_loaded = GLOBAL_MODULE

    settings_dict = {}
    for m in dir(mod):
        if m.isupper():
            # sets.add(m)
            value = getattr(mod, m)
            settings_dict[m] = value

    bp = settings_dict.get("BASE_PATH")
    if bp:
        base_path = bp
    else:
        settings_dict["BASE_PATH"] = base_path
    cfg = Settings(**settings_dict)
    cfg.SETTINGS_MODULE = module_loaded

    if not cfg.DEBUG:
        _level = _get_level(cfg.LOGLEVEL)
    else:
        _level = logging.DEBUG

    # set BASE_PATH
    # os.environ[defaults.BASE_PATH_ENV] = cfg.BASE_PATH
    # logging.config.dictConfig(cfg.LOGCONFIG)
    return cfg



import sys
from pathlib import Path

import click
from rich.console import Console

from services import conf, defaults, errors, types, utils
from services.security import jwtauth

# from services.security.memory_store import MemoryTokenStore

console = Console()


@click.group(name="tokens")
def tokens_cli():
    """
    Token generator
    """
    pass


@tokens_cli.command(name="generate")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option("--scopes", "-S", default="user:r:w", help="Scopes for the token")
@click.option("--exp", "-e", default=600, help="Expiration of the token")
@click.option("--username", "-u", default=None, help="Username")
def generate(settings_module, username, scopes, exp):
    """Generates a JWT token"""
    settings = conf.load_conf(settings_module)
    custom = {"scopes": scopes}
    if username:
        custom.update({"usr": username})
    payload = types.JWTPayload(custom=custom, exp=exp)

    jwt = jwtauth.JWTAuth(settings.SECURITY)
    token = utils.from_sync2async(jwt.generate_token, payload)

    console.print(token.dict())


@tokens_cli.command(name="decode")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("token")
def decode(settings_module, token):
    """Decodes a given token"""
    settings = conf.load_conf(settings_module)
    jwt = jwtauth.JWTAuth(settings.SECURITY)
    decoded = jwt.decode(token)

    console.print(decoded)


@tokens_cli.command(name="validate")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option("--scopes", "-S", default=None, help="Scopes for the token: user:r:w")
@click.argument("token")
def validate(settings_module, token, scopes):
    """Validates a given token"""
    settings = conf.load_conf(settings_module)
    jwt = jwtauth.JWTAuth(settings.SECURITY)
    try:
        _ = jwt.validate(token, required_scopes=scopes)
    except errors.AuthValidationFailed:
        console.print("[red bold]Invalid token[/]")
        sys.exit(-1)

    console.print("[green bold]Valid token[/]")


@tokens_cli.command(name="es512")
@click.argument("directory")
def es512(directory):
    """Generates the ES512 keys using openssl under the hood"""
    path = Path(directory)
    utils.mkdir_p(path)

    console.print(f">> private key into {path.resolve()}/private.key")
    utils.execute_cmd(
        f"openssl ecparam -name prime256v1 -genkey -out {path.resolve()}/private.key"
    )
    console.print(f">> public key into {path.resolve()}/public.key")
    try:
        utils.execute_cmd(
            f"openssl ec -in {path.resolve()}/private.key -pubout -out {path.resolve()}/public.key"
        )
    except errors.CommandExecutionException:
        pass



import os
import sys
from functools import partial
from typing import List

import click
from rich.console import Console
from rich.table import Table

from services import conf, defaults, types, workers, utils

console = Console()


async def _list_tasks(backend: types.TasksBackend) -> List[workers.Task]:
    back = await workers.init_backend(backend)
    tasks = await back.list_tasks()
    return tasks


async def _failed(backend: types.TasksBackend) -> List[str]:
    back = await workers.init_backend(backend)
    tasks = await back.clean_failed()
    return tasks


async def _delete(backend: types.TasksBackend, taskid: str):
    back = await workers.init_backend(backend)
    await back.delete_task(taskid)


async def _clean(backend: types.TasksBackend):
    back = await workers.init_backend(backend)
    await back.clean()


@click.group(name="tasks")
def tasks_cli():
    """
    Run list and configure tasks
    """
    pass


@tasks_cli.command(name="run")
@click.argument("taskname")
@click.option("--param", "-p", multiple=True, help="define a param as <key>=<value>")
@click.option("--param-from-json", "-j", help="load params from a json file")
@click.option("--param-from-yaml", "-y", help="load params from a yaml file")
@click.option("--timeout", "-t", default=60, help="timeout for the task")
@click.option("--worker-type", "-w", default="io", type=click.Choice(["io", "cpu"]))
@click.option("--package", required=True)
# @click.option("--backend-uri", "-u", default=None)
# @click.option("--backend-class", default="services.ext.sql.workers.SQLBackend")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def run_cli(
    taskname,
    param,
    worker_type,
    timeout,
    package,
    param_from_json,
    param_from_yaml,
    settings_module,
    # backend_uri,
    # backend_class,
):
    """Run a task"""

    settings = conf.load_conf(settings_module)
    if param_from_json:
        console.print("[red bold]json file params is not implemented yet[/]")
        sys.exit(-1)
    if param_from_yaml:
        console.print("[red bold]yaml file params is not implemented yet[/]")
        sys.exit(-1)
    qconf = workers.QueueConfig(app_name=package, backend=settings.TASKS)

    params = {}
    for p in param:
        k, v = p.split("=")
        params[k] = v

    task = workers.Task(name=taskname, timeout=timeout, params=params)
    if worker_type == "io":
        workers.standalone_io_worker(qconf, task)
    else:
        workers.standalone_cpu_worker(qconf, task)


@tasks_cli.command(name="list")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def list_tasks(settings_module):
    """List tasks"""
    settings = conf.load_conf(settings_module)
    if not settings.TASKS:
        console.print("[red bold]Not tasks backend configurated[/]")
        sys.exit(-1)

    table = Table(title="List of tasks")
    table.add_column("id")
    table.add_column("name")
    table.add_column("app_name")
    table.add_column("state")
    table.add_column("elapsed")

    tasks: List[workers.Task] = utils.from_sync2async(_list_tasks, settings.TASKS)
    for t in tasks:
        elapsed = round((t.updated_at - t.created_at).total_seconds())
        if elapsed > 120:
            elapsed = f"~{round(elapsed / 60)}m"
        else:
            elapsed = f"{elapsed}s"

        table.add_row(
            f"{t.id}",
            f"{t.name}",
            f"{t.app_name}",
            f"{t.state}",
            f"{elapsed}",
        )
    console.print(table)


@tasks_cli.command(name="clean-failed")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def clean_failed_tasks(settings_module):
    """Remove tasks in failed state"""
    settings = conf.load_conf(settings_module)
    if not settings.TASKS:
        console.print("[red bold]Not tasks backend configurated[/]")
        sys.exit(-1)

    tasks = utils.from_sync2async(_failed, settings.TASKS)
    for t in tasks:
        console.print(f"[red]=> {t} removed[/]")
    console.print(f"\n [bold]{len(tasks)}[/bold] failed tasks removed.")


@tasks_cli.command(name="delete")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("taskid")
def delete_task(settings_module, taskid):
    """Remove a tasks by taksid"""
    settings = conf.load_conf(settings_module)
    if not settings.TASKS:
        console.print("[red bold]Not tasks backend configurated[/]")
        sys.exit(-1)

    utils.from_sync2async(_delete, settings.TASKS, taskid)
    console.print(f"[red]=> {taskid} removed[/]")


@tasks_cli.command(name="clean")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def clean_tasks(settings_module):
    """Clean tasks done"""
    settings = conf.load_conf(settings_module)
    if not settings.TASKS:
        console.print("[red bold]Not tasks backend configurated[/]")
        sys.exit(-1)

    utils.from_sync2async(_clean, settings.TASKS)
    console.print(f"[green]=> Tasks cleaned[/]")



import importlib

import click
from sqlalchemy.sql.schema import MetaData

from services import conf, defaults, utils
from services.db import SQL, Migration
from services.db.helpers import delete_table
from services.db.ext import utils as dbutils
from rich.console import Console

console = Console()


def get_meta_from_app(app_name) -> MetaData:
    Base = utils.get_class(f"{app_name}.db.Base")
    return Base.metadata


@click.group(name="db")
def db_cli():
    """
    DB related tasks
    """
    pass


@db_cli.command(name="create-tables")
@click.option("--db", default="default", help="db name in the settings module")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def create_tables(app_name, db, settings_module):
    """Creates database schemas.

    Models should be accesible from {app_name}.models"""
    settings = conf.load_conf(settings_module)

    models_module = importlib.import_module(f"{app_name}.models")

    db = SQL.from_conf(settings.DATABASES[db])
    db.create_all(get_meta_from_app(app_name))


@db_cli.command("list-tables")
@click.option("--db", default="default", help="db name in the settings module")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def list_tables(app_name, db, settings_module):
    """List tables of a database"""

    settings = conf.load_conf(settings_module)

    _ = importlib.import_module(f"{app_name}.models")

    db = SQL.from_conf(settings.DATABASES[db])
    tables = db.list_tables()
    console.print(tables)


@db_cli.command("drop-table")
@click.option("--db", default="default", help="db name in the settings module")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option(
    "--table",
    "-t",
    help="Name of the table to drop",
)
@click.argument("app_name")
def drop_table(app_name, db, settings_module, table):
    """Delete a table"""

    settings = conf.load_conf(settings_module)

    _ = importlib.import_module(f"{app_name}.models")

    _db = SQL.from_conf(settings.DATABASES[db])
    res = delete_table(_db, table)
    if res:
        console.print(f"[green]Table {table} deleted[/]")
    else:
        console.print(f"[red]Delete of table {table} failed[/]")


@db_cli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def drop(app_name, db, settings_module):
    """drop database schemas"""
    settings = conf.load_conf(settings_module)

    models_module = importlib.import_module(f"{app_name}.models")

    db = SQL.from_conf(settings.DATABASES[db])
    db.drop_all(get_meta_from_app(app_name), all_=True)


@db_cli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini", help="alembic file")
@click.option("--to", "-t", default="head", help="Revision to upgrade or downgrade")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def upgrade(app_name, db, settings_module, alembic, to):
    """upgrade a database based on the migrations created. Use as:

    srv db upgrade example_app

    """
    settings = conf.load_conf(settings_module)

    m = Migration(
        settings.DATABASES[db].sync_url, section=app_name, alembic_file=alembic
    )
    m.upgrade(to=to)


@db_cli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini", help="alembic file")
@click.option("--to", "-t", default="head", help="Revision to upgrade or downgrade")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def downgrade(app_name, db, settings_module, alembic, to):
    """downgrade a database based on the migrations created"""
    settings = conf.load_conf(settings_module)

    m = Migration(
        settings.DATABASES[db].sync_url, section=app_name, alembic_file=alembic
    )
    m.downgrade(to=to)


@db_cli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini")
@click.option("--message", "-m", default=None, help="Revision name")
@click.option("--rev-id", "-R", help="Revision id")
@click.option(
    "--head",
    default=None,
    help="Specify head revision or <branchname>@head to base new revision.",
)
@click.option(
    "--branch-label",
    "-b",
    default=None,
    help="Specify a branch label to apply to the new revision",
)
@click.option("--depends-on", "-d", help="list of 'depends on' identifiers")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.argument("app_name")
def revision(
    app_name,
    db,
    alembic,
    message,
    rev_id,
    head,
    branch_label,
    depends_on,
    settings_module,
):
    """generates migrations files from models defined in apps. Use as:

    srv db revision example_app -m first -R 0001 -m first
    """

    settings = conf.load_conf(settings_module)

    m = Migration(
        settings.DATABASES[db].sync_url, section=app_name, alembic_file=alembic
    )

    m.revision(
        rev_id=rev_id,
        message=message,
        head=head,
        branch_label=branch_label,
        depends_on=depends_on,
        autogenerate=True,
    )


@db_cli.command(name="create-db")
@click.option("--db", default="default", help="db name in the settings module")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
@click.option(
    "--encoding",
    "-e",
    default="utf8",
    help="Encoding",
)
def create_db(db, settings_module, encoding):
    """Creates database"""

    settings = conf.load_conf(settings_module)

    _db = settings.DATABASES[db]
    dbutils.create_databese(_db.sync_url, encoding=encoding)



import os
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Generator,
    AsyncGenerator,
    List,
    Optional,
    Union,
)

import aiofiles

from services import types
from services.utils import from_async2sync, get_class, mkdir_p
from sanic.log import logger
from sanic import Sanic


def delete_file_or_dir(fpath):
    try:
        Path(fpath).unlink()
    except IsADirectoryError:
        Path(fpath).rmdir()


class KeyReadError(Exception):
    def __init__(self, bucket, key):
        msg = f"Value not found for key {key} in {bucket}"
        super().__init__(msg)


class KeyWriteError(Exception):
    def __init__(self, bucket, key):
        msg = f"Value not written for key {key} in {bucket}"
        super().__init__(msg)


class IStorage(ABC):
    """
    This is a generic KV store mostly use for project data related
    The interface is very simple allowing the common put/get actions
    either loading data in memory or streaming data for big files
    For examples about how to use some of them see tests/test_io_kv.py
    This interface is offered in a sync and async version
    """

    _bucket: str
    opts: Optional[Dict[str, Any]] = None

    @abstractmethod
    def get_bucket(self) -> types.Bucket:
        pass

    @abstractmethod
    def create_bucket(self, opts: Optional[Dict[str, str]] = None) -> types.Bucket:
        pass

    @abstractmethod
    def put(self, key: str, bdata: bytes):
        pass

    @abstractmethod
    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        pass

    @abstractmethod
    def delete(self, key: str):
        pass

    @abstractmethod
    def list(self) -> List[str]:
        pass

    @abstractmethod
    def get(self, key: str) -> Union[bytes, str, None]:
        pass

    @abstractmethod
    def get_stream(
        self, key: str, buffer_size=256 * 1024
    ) -> Generator[bytes, None, None]:
        pass

    @staticmethod
    def create(store_class, bucket, opts: Dict[str, Any] = {}) -> "IStorage":
        Class = get_class(store_class)
        return Class(bucket, opts)

    @abstractmethod
    def download_signed(self, key, minutes=15, bucket=None) -> str:
        pass

    @abstractmethod
    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        pass


class IAsyncStore(ABC):
    """
    This is a generic KV store mostly use for project data related
    The interface is very simple allowing the common put/get actions
    either loading data in memory or streaming data for big files
    For examples about how to use some of them see tests/test_io_kv.py
    This interface is offered in a sync and async version
    """

    _bucket: str
    opts: Optional[Dict[str, Any]] = None

    @abstractmethod
    async def create_bucket(
        self, opts: Optional[Dict[str, str]] = None
    ) -> types.Bucket:
        pass

    @abstractmethod
    async def get_bucket(self) -> types.Bucket:
        pass

    @classmethod
    @abstractmethod
    async def from_uri(
        cls, bucket, opts: Optional[Dict[str, Any]] = None
    ) -> "IAsyncStore":
        pass

    @abstractmethod
    async def put(self, key: str, bdata: bytes):
        pass

    @abstractmethod
    async def put_stream(self, key: str, generator: AsyncIterator) -> bool:
        pass

    @abstractmethod
    async def get(self, key: str) -> Union[bytes, str, None]:
        pass

    @abstractmethod
    async def list(self) -> List[str]:
        pass

    @abstractmethod
    async def delete(self, key: str):
        pass

    @abstractmethod
    async def get_stream(self, key: str, buffer_size=1024):
        pass

    @abstractmethod
    def download_signed(self, key, minutes=15, bucket=None) -> str:
        pass

    @abstractmethod
    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        pass


class Local(IStorage):
    def __init__(self, bucket: str):
        self._opts = None
        self._bucket = bucket

    def create_bucket(self):
        mkdir_p(self._bucket)
        return types.Bucket(
            name=self._bucket,
            url=self._bucket,
            storage_class="filesystem",
            location="local",
            versioning=False,
            created_at=datetime.utcnow(),
        )

    def get_bucket(self) -> types.Bucket:
        return types.Bucket(
            name=self._bucket,
            url=self._bucket,
            storage_class="filesystem",
            location="local",
            versioning=False,
            created_at=datetime.utcnow(),
        )

    def uri(self, key):
        return f"{self._bucket}/{key}"

    def put(self, key: str, bdata: bytes):
        # obj = io.BytesIO(bdata)
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            with open(uri, "wb") as f:
                f.write(bdata)
        except Exception as e:
            raise KeyWriteError(self._bucket, key) from e

    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            with open(uri, "wb") as f:
                for chunk in generator:
                    f.write(chunk)
        except Exception as e:
            raise KeyWriteError(self._bucket, key) from e

        return True

    def get(self, key: str) -> Union[bytes, None]:
        uri = self.uri(key)
        try:
            with open(uri, "rb") as f:
                obj = f.read()
                return obj
        except Exception as e:
            raise KeyReadError(self._bucket, key) from e

    def _read_chunks(self, fobject, chunk_size=1024):
        while True:
            data = fobject.read(chunk_size)
            if not data:
                break
            yield data

    def get_stream(self, key: str, buffer_size=1024) -> Generator[bytes, None, None]:
        uri = self.uri(key)
        try:
            with open(uri, "rb") as f:
                for chunk in self._read_chunks(f, buffer_size):
                    yield chunk
        except Exception as e:
            raise KeyReadError(self._bucket, key) from e

    def list(self) -> List[str]:
        return os.listdir(self._bucket)

    def delete(self, key: str):
        delete_file_or_dir(f"{self._bucket}/{key}")

    def download_signed(self, key, minutes=15, bucket=None) -> str:
        # _to_sign = f"{self._bucket}/{key}:valid={minutes}"
        # uri = self._sign.dumps(_to_sign)
        # return f"{self._bucket}/{key}?key={uri}"
        raise NotImplementedError()

    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        raise NotImplementedError()


class AsyncLocal(IAsyncStore):
    """For local usage and testing"""

    def __init__(self, bucket: str):
        self.opts = None
        self._bucket = bucket
        # self._sign = URLSafeTimedSerializer(
        #     secret_key=client_opts.get("secret_key"), salt=b"storage"
        # )

    def uri(self, key):
        return f"{self._bucket}/{key}"

    @classmethod
    async def from_uri(
        cls, bucket, opts: Optional[Dict[str, Any]] = None
    ) -> "IAsyncStore":
        obj = cls(bucket)

        return obj

    async def create_bucket(
        self, opts: Optional[Dict[str, str]] = None
    ) -> types.Bucket:
        await from_async2sync(mkdir_p, self._bucket)
        return types.Bucket(
            name=self._bucket,
            url=self._bucket,
            storage_class="filesystem",
            location="local",
            versioning=False,
            created_at=datetime.utcnow(),
        )

    async def get_bucket(self) -> types.Bucket:
        return types.Bucket(
            name=self._bucket,
            url=self._bucket,
            storage_class="filesystem",
            location="local",
            versioning=False,
            created_at=datetime.utcnow(),
        )

    async def put(self, key: str, bdata: bytes):
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            async with aiofiles.open(uri, mode="wb") as f:
                await f.write(bdata)
        except Exception as e:
            raise KeyWriteError(self._bucket, key) from e

    async def put_stream(self, key: str, generator: AsyncIterator) -> bool:
        uri = self.uri(key)
        # mkdir_p((Path(uri).parent).resolve())
        try:
            async with aiofiles.open(uri, mode="wb") as f:
                async for data in generator:
                    await f.write(data)
        except Exception as e:
            raise KeyWriteError(self._bucket, key) from e

        return True

    async def get(self, key: str) -> Union[bytes, str, None]:
        uri = self.uri(key)
        try:
            async with aiofiles.open(uri, mode="rb") as f:
                data = await f.read()
                return data
        except Exception as e:
            raise KeyReadError(self._bucket, key) from e

    async def get_stream(self, key: str, buffer_size=1024):
        """PEP 0525 for Asynchronous generators"""
        uri = self.uri(key)
        try:
            async with aiofiles.open(uri, mode="rb") as f:
                while True:
                    data = await f.read(buffer_size)
                    if not data:
                        break
                    yield data

        except Exception as e:
            raise KeyReadError(self._bucket, key) from e

    async def list(self) -> List[str]:
        rsp = await from_async2sync(os.listdir, self._bucket)
        return rsp

    async def delete(self, key: str):
        await from_async2sync(delete_file_or_dir, f"{self._bucket}/{key}")

    def download_signed(self, key, minutes=15, bucket=None) -> str:
        # _to_sign = f"{self._bucket}/{key}:valid={minutes}"
        # uri = self._sign.dumps(_to_sign)
        # return f"{self._bucket}/{key}?key={uri}"
        raise NotImplementedError()

    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        raise NotImplementedError()


class Storage:
    def __init__(self, app: Sanic, settings: types.Settings, create_bucket=False):
        self.app = app
        app.config.STORAGE = settings.STORAGE
        self._create_bucket = create_bucket
        # app.ext.dependency(self)

    async def _init_store(self, bucket: str, store_class: str) -> IAsyncStore:
        _Store: IAsyncStore = get_class(store_class)
        obj: IAsyncStore = await _Store.from_uri(bucket)
        if self._create_bucket:
            await obj.create_bucket()
        return obj

    async def init_listener(self, app: Sanic):
        app.ctx.storage = {}
        for k, v in app.config.STORAGE.items():
            obj = await self._init_store(v.bucket, v.store_class)
            app.ctx.storage[k] = obj
            logger.info("%s storage added", k)

    # @classmethod
    # def init_app(cls, app, settings: types.Settings) -> "Storage":
    #     obj = cls(app, settings)
    #     app.register_listener(obj.init_listener, "after_server_start")
    #     app.ext.dependency(obj)
    #     return obj

    def get_storage(self, name="default") -> IAsyncStore:
        return self.app.ctx.storage[name]


def init_app(app, settings: types.Settings, create_bucket=False):
    s = Storage(app, settings, create_bucket=create_bucket)
    app.register_listener(s.init_listener, "after_server_start")
    app.ext.dependency(s)



import pytest
import pytest_asyncio
from services import types, utils
from services.db import AsyncSQL, SQL
from services.db.helpers import drop_everything
from tests.common import async_create_all, create_all, Base

_db_name = "test.db"
testdb = types.Database(
    name="testing",
    async_url=f"sqlite+aiosqlite:///{_db_name}",
    sync_url=f"sqlite:///{_db_name}",
    description="Dataproc database",
)


@pytest_asyncio.fixture()
async def adb():
    _db = AsyncSQL.from_conf(testdb)

    await async_create_all(_db)

    yield _db
    # await utils.from_async2sync(drop_everything, _db.engine)
    await _db.drop_all(Base.metadata, all_=False)

    # await _db.drop_all(Base.metadata)
    # await _db.drop_all()


@pytest.fixture(scope="module")
def db():
    _db = SQL.from_conf(testdb)
    create_all(_db)
    yield _db
    _db.drop_all()
    # await _db.drop_all(Base.metadata)



from importlib import import_module
from typing import Callable

from sanic import Request, Sanic
from sanic.response import json

from services import defaults, storage
from services.base import WebAppSpec
from services.db.web import init_db
from services.redis_conn import create_pool
from services.templates import Render
from services.types import Settings
from services.utils import get_class, get_version

version = get_version()


async def stream_reader(request: Request):
    """
    It's a wrapper to be used to yield response from a stream
    to another stream.
    it's used with project upload data to stream upload zip directly to
    the fileserver instead of load data in memory.
    """
    while True:
        body = await request.stream.read()
        if body is None:
            break
        yield body


async def status_handler(request):
    return json(dict(msg="We are ok", version=version))


def create_srv(
    settings: Settings,
    app_name=defaults.SANIC_APP,
    create_redis=create_pool,
    with_status_handler=True,
    auth_enabled=True,
    init_db_func: Callable = init_db,
) -> Sanic:
    """Factory pattern like flask"""

    app = Sanic(app_name)

    app.config.CORS_ORIGINS = settings.CORS_ORIGINS
    app.config.OAS_UI_DEFAULT = "swagger"
    app.config.OAS_UI_REDOC = False
    app.config.SWAGGER_UI_CONFIGURATION = {
        "apisSorter": "alpha",
        "operationsSorter": "alpha",
        "docExpansion": "none",
    }
    app.config.SETTINGS = settings
    if settings.REDIS:
        app.ctx.redis = create_redis(settings.REDIS)
    if settings.DATABASES:
        init_db_func(app, settings=settings)

    if auth_enabled:
        app.ext.openapi.add_security_scheme(
            "token",
            "http",
            scheme="bearer",
            bearer_format="JWT",
        )
        app.ext.openapi.secured("token")
        # secure_app: WebAppSpec = get_class("services.security.web.WebApp")()
        # secure_app.init(app, settings)
        # Authenticator(app)

    for sd in settings.STATICFILES_DIRS:
        app.static(sd.uripath, sd.localdir, name=sd.name)

    render: Render = Render(searchpath=settings.TEMPLATES_DIR)
    render.init_app(app)
    render.add_static(app, settings)
    if settings.VITE_ENABLED:
        render.add_vite(app, settings)

    app.ctx.databases = {}
    if settings.DATABASES:
        init_db_func(app, settings)

    for wapp in settings.APPS:
        WebAppClass: WebAppSpec = get_class(wapp)
        WebAppClass(app, settings)

    if with_status_handler:
        app.add_route(status_handler, "/status", name="status")

    if settings.STORAGE:
        storage.init_app(app, settings)

    return app



import importlib
import os
from logging.config import fileConfig

from {{data.app_name}}.db import Base
from sqlalchemy import engine_from_config, pool

from alembic import context


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
models_mod = config.get_main_option("models_module")
models = importlib.import_module(models_mod)
# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def include_name(name, type_, parent_names):
    if type_ == "table":
        return name in target_metadata.tables
    return True


def run_migrations_offline():
    """Run migrations in 'offline' mode.
    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.
    Calls to context.execute() here emit the given string to the
    script output.
    """
    url = config.get_main_option("sqlalchemy.url")
    url = url or os.getenv("SRV_SQL")
    version_table = config.get_main_option("version_table")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        include_name=include_name,
        dialect_opts={"paramstyle": "named"},
        version_table=version_table,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.
    In this scenario we need to create an Engine
    and associate a connection with the context.
    """

    cfg = config.get_section(config.config_ini_section)
    url = config.get_main_option("sqlalchemy.url")
    url = url or os.getenv("SRV_SQL")
    version_table = config.get_main_option("version_table")

    connectable = engine_from_config(
        cfg,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=version_table,
            include_name=include_name,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from typing import Any, Dict

from sqlalchemy import BigInteger, Boolean, Column, DateTime, String

# from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.orm import declarative_mixin, declared_attr
from sqlalchemy.sql import functions
from sqlalchemy.types import BINARY


@declarative_mixin
class UserMixin:
    id = Column(BigInteger, primary_key=True)
    username = Column(String(), index=True, unique=True, nullable=False)
    password = Column("password", BINARY, nullable=True)
    email = Column(String(), index=True, nullable=True, unique=True)
    is_superuser = Column(Boolean, default=False, nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)
    scopes = Column(String(), default="user", nullable=False)

    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError()

    def __repr__(self):
        return f"<UserModel: {self.id}:{self.username}>"


@declarative_mixin
class GroupMixin:
    id = Column(BigInteger, primary_key=True)
    group_name = Column(String(16), unique=True, nullable=False, index=True)
    is_active = Column(Boolean, default=True, nullable=False)
    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )

    def __repr__(self):
        return f"<GroupModel: {self.id}:{self.group_name}>"



import click
from services.db import AsyncSQL
from services.conf import load_conf
from services import defaults


@click.command(name="shell")
@click.option(
    "--settings-module",
    "-s",
    default=defaults.SETTINGS_MODULE,
    help="Fullpath to settings module",
)
def shell_cli(settings_module):
    """starts a IPython REPL console with db objects and models"""
    try:
        from IPython import start_ipython
    except ImportError:
        print("You need to install ipython:\n\t pip install ipython")

    settings = load_conf(settings_module)

    db = AsyncSQL.from_conf(settings.DATABASES["default"])
    # S = db.sessionmaker()
    start_ipython(
        argv=[],
        user_ns={
            "settings": settings,
            # "Session": S,
            "db": db,
        },
    )



from datetime import datetime
from typing import Any, Dict, Union

from services import types
from services.db.managers import AsyncManagerBase, AsyncSecretBase, SecretBase
from services.errors import (
    AuthValidationFailed,
    DBObjectNotFound,
    GroupNotFound,
    UserNotFound,
)
from services.security import PasswordScript

from .models import SecretsModel
{% if data.users -%}
from .users_models import GroupModel, UserModel


class GroupManager(AsyncManagerBase[GroupModel]):
    model_class = "{{ data.app_name }}.users_models.GroupModel"
    lookup_key = "group_name"

    async def assign_user(self, session, group_name, user: UserModel):
        group = await self.get(session, group_name)
        group.users.append(user)
        session.add(group)


class UserManager(AsyncManagerBase[UserModel]):
    model_class = "{{ data.app_name }}.users_models.UserModel"
    lookup_key = "username"

    def __init__(self, salt: str, groups: GroupManager):
        super().__init__()
        self.groups = groups
        self._salt = salt

    async def assign_group(self, session, group_name: str, user: UserModel):
        await self.groups.assign_user(session, group_name, user)

    async def register(
        self, session, data: Dict[str, Any], default_group="users"
    ) -> Union[UserModel, None]:
        _copy = data.copy()
        _copy["password"] = self.encrypt_password(_copy["password"], self._salt)

        um = await self.create(session, _copy, commit=False)
        if default_group and um:
            try:
                await self.groups.assign_user(session, default_group, um)
            except DBObjectNotFound:
                raise GroupNotFound(default_group)

        res = await self._commit_or_rollback(session)
        if not res:
            return None
        return um

    async def create_group(self, session, group_name) -> GroupModel:
        return await self.groups.create(session, {"group_name": group_name})

    async def authenticate(
        self, session, *, username: str, to_verify: str
    ) -> UserModel:
        try:
            user = await self.get(session, username)
            is_valid = self.verify_password(
                user.password, to_verify=to_verify, salt=self._salt
            )
            if not is_valid:
                raise AuthValidationFailed()
            return user

        except DBObjectNotFound:
            raise AuthValidationFailed()

    def verify_password(
        self, original_password: bytes, *, to_verify: str, salt: Union[bytes, str]
    ) -> bool:
        pm = PasswordScript(salt)
        verified = pm.verify(to_verify, original_password)
        if verified:
            return True
        return False

    def encrypt_password(self, password, salt: Union[bytes, str]) -> bytes:
        pm = PasswordScript(salt)
        encrypted = pm.encrypt(password)
        return encrypted

    async def change_pass(self, session, username: str, new_password: str):
        u = await self._get_one(session, username)
        if not u:
            raise UserNotFound(username)
        pass_ = self.encrypt_password(new_password, salt=self._salt)
        u.password = pass_
        u.updated_at = datetime.utcnow()
        session.add(u)

{% endif -%}

class AsyncSecrets(AsyncSecretBase[SecretsModel]):
    def __init__(self, settings: types.Settings):
        super().__init__(private_key=settings.SECURITY.secret_key)


class Secrets(SecretBase[SecretsModel]):
    def __init__(self, settings: types.Settings):
        super().__init__(private_key=settings.SECURITY.secret_key)



import asyncio
import codecs
import importlib.util
import json
import os
import re
import secrets
import subprocess
import unicodedata
from enum import Enum
from importlib import import_module
from pathlib import Path
from typing import Callable, Union

import aiofiles
import click

from services import __about__
from services.errors import CommandExecutionException

# from sanic_ext import openapi


_filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
_windows_device_files = (
    "CON",
    "AUX",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "LPT1",
    "LPT2",
    "LPT3",
    "PRN",
    "NUL",
)


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path=None):
    # if rel_path:
    #     for line in read(rel_path).splitlines():
    #         if line.startswith("__version__"):
    #             delim = '"' if '"' in line else "'"
    #             return line.split(delim)[1]
    return __about__.__version__


def get_query_param(request, key, default_val=None):
    val = request.args.get(key, default_val)
    return val


def mkdir_p(fp):
    """Make the fullpath
    similar to mkdir -p in unix systems.
    """
    Path(fp).mkdir(parents=True, exist_ok=True)


def open_json(path):
    with open(path, "r", encoding="utf-8") as f:
        data = f.read()
        dict_ = json.loads(data)
    return dict_


def execute_cmd(cmd) -> str:
    """Wrapper around subprocess"""
    with subprocess.Popen(
        cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:
        out, err = p.communicate()
        if err:
            raise CommandExecutionException(err.decode())

        return out.decode().strip()


def path_norm(fp):
    """Given a  filepath returns a normalized a path"""
    return str(Path(fp))


def get_parent_folder():
    """Get only the name of the parent folder
    commonly used to define the project name
    """
    root = Path(os.getcwd())
    return str(root).rsplit("/", maxsplit=1)[-1]


def get_package_dir(pkg):
    spec = importlib.util.find_spec(pkg)
    return spec.submodule_search_locations[0]


def get_from_module(fullpath):
    splited = fullpath.rsplit(".", maxsplit=1)
    mod_name = splited[0]
    obj_name = splited[1]
    mod = import_module(mod_name)
    obj = getattr(mod, obj_name)
    return obj


def secure_filename(filename: str) -> str:
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.
    On windows systems the function also makes sure that the file is not
    named after one of the special device files.
    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename('i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'
    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you abort or
    generate a random filename if the function returned an empty one.
    .. versionadded:: 0.5
    :param filename: the filename to secure
    """
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.encode("ascii", "ignore").decode("ascii")

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip(
        "._"
    )

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if (
        os.name == "nt"
        and filename
        and filename.split(".")[0].upper() in _windows_device_files
    ):
        filename = f"_{filename}"

    return filename


def normalize_name(name: str) -> str:
    """used mostly for projects"""
    evaluate = name.lower()
    evaluate = evaluate.replace(" ", "_")
    evaluate = secure_filename(name)
    return evaluate


def get_class(fullclass_path):
    """get a class or object from a module. The fullclass_path should be passed as:
    package.my_module.MyClass
    """
    module, class_ = fullclass_path.rsplit(".", maxsplit=1)
    mod = import_module(module)
    cls = getattr(mod, class_)
    return cls


def init_blueprints_legacy(app, blueprints_allowed, package_dir="services.web"):
    """
    It will import bluprints from modules that ends with "_bp" and belongs
    to the package declared in `package_dir`
    """
    blueprints = set()
    mod = app.__module__
    for mod_name in blueprints_allowed:
        modules = import_module(f"{package_dir}.{mod_name}", mod)
        for el in dir(modules):
            if el.endswith("_bp"):
                bp = getattr(modules, el)
                blueprints.add(bp)

    for bp in blueprints:
        print("Adding blueprint: ", bp.name)
        app.blueprint(bp)


class MimeTypes(Enum):
    """common mime types used
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types
    """

    html = "text/html"
    javascript = "text/javascript"
    json = "application/json"
    gz = "application/gzip"
    jpg = "image/jpeg"
    bz2 = "application/x-bzip2"
    css = "text/css"
    csv = "text/csv"
    doc = "application/msword"
    docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    gif = "image/gif"


def get_function(fullname) -> Callable:
    mod, name = fullname.rsplit(".", maxsplit=1)
    pkg = mod.split(".", maxsplit=1)[0]
    try:
        module = import_module(mod, pkg)
    except (ModuleNotFoundError, AttributeError):
        raise KeyError(fullname)
    return getattr(module, name)


def secure_random_str(size=12) -> str:
    return secrets.token_urlsafe(size)


async def from_async2sync(func, *args, **kwargs):
    """Run sync functions from async code"""
    loop = asyncio.get_running_loop()
    rsp = await loop.run_in_executor(None, func, *args, **kwargs)
    return rsp


def from_sync2async(func, *args, **kwargs):
    """run async functions from sync code"""
    loop = asyncio.get_event_loop()
    rsp = loop.run_until_complete(func(*args, **kwargs))
    return rsp


def binary_file_reader(fp: str, chunk_size=1024):
    """
    File reader generator mostly used for projects upload
    """
    with open(fp, "rb") as f:
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            yield data


async def async_binary_reader(fp: str, chunk_size=1024):
    """
    It should be used as:
    async for chunk in async_binary_reader(fp):
        print(chunk)
    """
    async with aiofiles.open(fp, "rb") as f:
        while True:
            data = await f.read(chunk_size)
            if not data:
                break
            yield data


def to_schema(model_def):
    """
    Based on
    https://github.com/sanic-org/sanic-ext/issues/152
    It converts a pydantic model into a compatible schema for openapi
    """
    # openapi.component(model_def)
    return model_def.schema(ref_template="#/components/schemas/{model}")



import sys
from importlib import import_module
from pathlib import Path
from typing import Union

import click
from rich.console import Console

from services import conf, defaults

# from services.commands.common import create_app
# from services.commands.db import dbcli
# from services.commands.shell import shellcli
# from services.commands.users import userscli
# from services.commands.web import webcli


def get_command(fullname_path) -> Union[click.core.Command, click.core.Group]:
    module, class_ = fullname_path.rsplit(".", maxsplit=1)
    mod = import_module(module)
    cls = getattr(mod, class_)
    return cls


def import_cli(cli_app, package_dir="services.commands"):
    mod = import_module(package_dir)
    for el in dir(mod):
        if el.endswith("_cli"):
            cmd = getattr(mod, el)
            cli_app.add_command(cmd)


def from_path_to_module_str(fp) -> str:
    """experimental:
    from "examples/model.py" it should return
    "example.model"
    """
    return fp.rsplit(".", maxsplit=1)[0].replace("/", ".")


def import_cli_from_current(cli_app):
    sys.path.append(str(Path(".")))
    for path in Path(".").rglob("commands/*.py"):
        if path.is_file():
            if str(path).split("/")[1] == "commands":
                _module = from_path_to_module_str(str(path))
                # mod = import_module(_module)
                import_cli(cli_app, _module)


def init_cli():
    console = Console()

    @click.group()
    def cli():
        """
        Command line tool
        """

    @click.command()
    def version():
        """Actual version"""
        from services.utils import get_version

        ver = get_version()
        console.print(f"[bold magenta]{ver}[/bold magenta]")

    # import_cli(cli)
    # import_cli_from_current(cli)
    cli.add_command(version)
    _conf = conf.load_conf(default_module=defaults.SETTINGS_MODULE)
    for cmd in _conf.COMMANDS:
        _cmd = get_command(cmd)
        cli.add_command(_cmd)

    return cli


cli = init_cli()

if __name__ == "__main__":
    cli()



from sanic.request import Request

from services import conf, defaults, types
from services.db import SQL, AsyncSQL
from services.templates import Render


async def async_render(request: Request, tpl_name, data) -> str:
    """ """
    render: Render = request.app.ctx.render
    tpl = await render.async_render(request, tpl_name, data=data)
    return tpl
    # return html(tpl, status=200)


def get_async_db(
    dbname="default", settings_module=defaults.SETTINGS_MODULE
) -> AsyncSQL:
    settings = conf.load_conf(settings_module)
    _db = AsyncSQL.from_conf(settings.DATABASES[dbname])
    return _db


def get_db(dbname="default", settings_module=defaults.SETTINGS_MODULE) -> SQL:
    settings = conf.load_conf(settings_module)
    _db = SQL.from_conf(settings.DATABASES[dbname])
    return _db


def load_conf(settings_module=defaults.SETTINGS_MODULE) -> types.Settings:
    _c = conf.load_conf(settings_module)
    return _c



# taken from
# https://sqlalchemy-utils.readthedocs.io/en/latest/_modules/sqlalchemy_utils/functions/database.html#create_database
import os
from copy import copy

import sqlalchemy as sa
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm.exc import UnmappedInstanceError
from sqlalchemy.orm.session import object_session


def _get_scalar_result(engine, sql):
    with engine.connect() as conn:
        return conn.scalar(sql)


def _set_url_database(url: sa.engine.url.URL, database):
    """Set the database of an engine URL.

    :param url: A SQLAlchemy engine URL.
    :param database: New database to set.

    """
    if hasattr(url, "_replace"):
        # Cannot use URL.set() as database may need to be set to None.
        ret = url._replace(database=database)
    else:  # SQLAlchemy <1.4
        url = copy(url)
        url.database = database
        ret = url
    assert ret.database == database, ret
    return ret


def _sqlite_file_exists(database):
    if not os.path.isfile(database) or os.path.getsize(database) < 100:
        return False

    with open(database, "rb") as f:
        header = f.read(100)

    return header[:16] == b"SQLite format 3\x00"


def get_bind(obj):
    """
    Return the bind for given SQLAlchemy Engine / Connection / declarative
    model object.

    :param obj: SQLAlchemy Engine / Connection / declarative model object

    ::

        from sqlalchemy_utils import get_bind


        get_bind(session)  # Connection object

        get_bind(user)

    """
    if hasattr(obj, "bind"):
        conn = obj.bind
    else:
        try:
            conn = object_session(obj).bind
        except UnmappedInstanceError:
            conn = obj

    if not hasattr(conn, "execute"):
        raise TypeError(
            "This method accepts only Session, Engine, Connection and "
            "declarative model objects."
        )
    return conn


def quote(mixed, ident):
    """
    Conditionally quote an identifier.
    ::


        from sqlalchemy_utils import quote


        engine = create_engine('sqlite:///:memory:')

        quote(engine, 'order')
        # '"order"'

        quote(engine, 'some_other_identifier')
        # 'some_other_identifier'


    :param mixed: SQLAlchemy Session / Connection / Engine / Dialect object.
    :param ident: identifier to conditionally quote
    """
    if isinstance(mixed, Dialect):
        dialect = mixed
    else:
        dialect = get_bind(mixed).dialect
    return dialect.preparer(dialect).quote(ident)


def database_exists(url):
    """Check if a database exists.

    :param url: A SQLAlchemy engine URL.

    Performs backend-specific testing to quickly determine if a database
    exists on the server. ::

        database_exists('postgresql://postgres@localhost/name')  #=> False
        create_database('postgresql://postgres@localhost/name')
        database_exists('postgresql://postgres@localhost/name')  #=> True

    Supports checking against a constructed URL as well. ::

        engine = create_engine('postgresql://postgres@localhost/name')
        database_exists(engine.url)  #=> False
        create_database(engine.url)
        database_exists(engine.url)  #=> True

    """

    url = make_url(url)
    database = url.database
    dialect_name = url.get_dialect().name
    engine = None
    try:
        if dialect_name == "postgresql":
            text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
            for db in (database, "postgres", "template1", "template0", None):
                url = _set_url_database(url, database=db)
                engine = sa.create_engine(url)
                try:
                    return bool(_get_scalar_result(engine, sa.text(text)))
                except (ProgrammingError, OperationalError):
                    pass
            return False

        elif dialect_name == "mysql":
            url = _set_url_database(url, database=None)
            engine = sa.create_engine(url)
            text = (
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = '%s'" % database
            )
            return bool(_get_scalar_result(engine, sa.text(text)))

        elif dialect_name == "sqlite":
            url = _set_url_database(url, database=None)
            engine = sa.create_engine(url)
            if database:
                return database == ":memory:" or _sqlite_file_exists(database)
            else:
                # The default SQLAlchemy database is in memory, and :memory: is
                # not required, thus we should support that use case.
                return True
        else:
            text = "SELECT 1"
            try:
                engine = sa.create_engine(url)
                return bool(_get_scalar_result(engine, sa.text(text)))
            except (ProgrammingError, OperationalError):
                return False
    finally:
        if engine:
            engine.dispose()


def create_database(url, encoding="utf8", template=None):
    """Issue the appropriate CREATE DATABASE statement.

    :param url: A SQLAlchemy engine URL.
    :param encoding: The encoding to create the database as.
    :param template:
        The name of the template from which to create the new database. At the
        moment only supported by PostgreSQL driver.

    To create a database, you can pass a simple URL that would have
    been passed to ``create_engine``. ::

        create_database('postgresql://postgres@localhost/name')

    You may also pass the url from an existing engine. ::

        create_database(engine.url)

    Has full support for mysql, postgres, and sqlite. In theory,
    other database engines should be supported.
    """

    url = make_url(url)
    database = url.database
    dialect_name = url.get_dialect().name
    dialect_driver = url.get_dialect().driver

    if dialect_name == "postgresql":
        url = _set_url_database(url, database="postgres")
    elif dialect_name == "mssql":
        url = _set_url_database(url, database="master")
    elif dialect_name == "cockroachdb":
        url = _set_url_database(url, database="defaultdb")
    elif not dialect_name == "sqlite":
        url = _set_url_database(url, database=None)

    if (dialect_name == "mssql" and dialect_driver in {"pymssql", "pyodbc"}) or (
        dialect_name == "postgresql"
        and dialect_driver
        in {"asyncpg", "pg8000", "psycopg", "psycopg2", "psycopg2cffi"}
    ):
        engine = sa.create_engine(url, isolation_level="AUTOCOMMIT")
    else:
        engine = sa.create_engine(url)

    if dialect_name == "postgresql":
        if not template:
            template = "template1"

        with engine.begin() as conn:
            text = "CREATE DATABASE {} ENCODING '{}' TEMPLATE {}".format(
                quote(conn, database), encoding, quote(conn, template)
            )
            conn.execute(sa.text(text))

    elif dialect_name == "mysql":
        with engine.begin() as conn:
            text = "CREATE DATABASE {} CHARACTER SET = '{}'".format(
                quote(conn, database), encoding
            )
            conn.execute(sa.text(text))

    elif dialect_name == "sqlite" and database != ":memory:":
        if database:
            with engine.begin() as conn:
                conn.execute(sa.text("CREATE TABLE DB(id int)"))
                conn.execute(sa.text("DROP TABLE DB"))

    else:
        with engine.begin() as conn:
            text = f"CREATE DATABASE {quote(conn, database)}"
            conn.execute(sa.text(text))

    engine.dispose()


def drop_database(url):
    """Issue the appropriate DROP DATABASE statement.

    :param url: A SQLAlchemy engine URL.

    Works similar to the :ref:`create_database` method in that both url text
    and a constructed url are accepted. ::

        drop_database('postgresql://postgres@localhost/name')
        drop_database(engine.url)

    """

    url = make_url(url)
    database = url.database
    dialect_name = url.get_dialect().name
    dialect_driver = url.get_dialect().driver

    if dialect_name == "postgresql":
        url = _set_url_database(url, database="postgres")
    elif dialect_name == "mssql":
        url = _set_url_database(url, database="master")
    elif dialect_name == "cockroachdb":
        url = _set_url_database(url, database="defaultdb")
    elif not dialect_name == "sqlite":
        url = _set_url_database(url, database=None)

    if dialect_name == "mssql" and dialect_driver in {"pymssql", "pyodbc"}:
        engine = sa.create_engine(url, connect_args={"autocommit": True})
    elif dialect_name == "postgresql" and dialect_driver in {
        "asyncpg",
        "pg8000",
        "psycopg",
        "psycopg2",
        "psycopg2cffi",
    }:
        engine = sa.create_engine(url, isolation_level="AUTOCOMMIT")
    else:
        engine = sa.create_engine(url)

    if dialect_name == "sqlite" and database != ":memory:":
        if database:
            os.remove(database)
    elif dialect_name == "postgresql":
        with engine.begin() as conn:
            # Disconnect all users from the database we are dropping.
            version = conn.dialect.server_version_info
            pid_column = "pid" if (version >= (9, 2)) else "procpid"
            text = """
            SELECT pg_terminate_backend(pg_stat_activity.{pid_column})
            FROM pg_stat_activity
            WHERE pg_stat_activity.datname = '{database}'
            AND {pid_column} <> pg_backend_pid();
            """.format(
                pid_column=pid_column, database=database
            )
            conn.execute(sa.text(text))

            # Drop the database.
            text = f"DROP DATABASE {quote(conn, database)}"
            conn.execute(sa.text(text))
    else:
        with engine.begin() as conn:
            text = f"DROP DATABASE {quote(conn, database)}"
            conn.execute(sa.text(text))

    engine.dispose()



import contextlib
from datetime import datetime
from functools import partial
from typing import Any, Dict, List

from sqlalchemy import (
    JSON,
    Column,
    DateTime,
    Integer,
    MetaData,
    String,
    Table,
)
from sqlalchemy import delete as sqldelete
from sqlalchemy import select, update
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.sql import functions

from services.db import async_set_pragma, async_vacuum

# from services.db.utils import CreateTableIfNotExists
from services.workers import (
    IState,
    Task,
    TaskStatus,
    elapsed_time_from_finish,
    elapsed_time_from_start,
)


class SQLBackend(IState):
    def __init__(
        self,
        engine: AsyncEngine,
        meta=MetaData(),
        table_state="tasks_state",
        # table_history="tasks_history",
    ):
        self.meta = meta
        self._tasks = self._create_tasks_table(table_state)
        # self._history = self._create_tasks_table(table_history)
        self.engine = engine

    async def _dispose(self):
        await self.engine.dispose()

    @contextlib.asynccontextmanager
    async def conn(self):
        try:
            async with self.engine.connect() as conn:
                yield conn
        finally:
            await self.engine.dispose()

    @contextlib.asynccontextmanager
    async def begin(self):
        try:
            async with self.engine.begin() as conn:
                yield conn
        finally:
            await self.engine.dispose()

    async def create_all(self):
        # f = partial(CreateTableIfNotExists, self._tasks)
        f = partial(self.meta.create_all, checkfirst=True)
        async with self.engine.begin() as conn:
            # await conn.run_sync(self.meta.create_all)
            try:
                await conn.run_sync(f)
            except OperationalError:
                pass

    @classmethod
    async def from_uri(cls, uri: str, extra: Dict[str, Any] = {}) -> "IState":
        _echo = extra.get("echo", False)
        _table = extra.get("table_state", "tasks_state")
        _wal = extra.get("wal", True)
        engine = create_async_engine(uri, echo=_echo)
        if _wal:
            await async_set_pragma(engine)

        obj = cls(engine, table_state=_table)
        await obj.create_all()
        return obj

    def _create_tasks_table(self, name):
        tbl = Table(
            name,
            self.meta,
            Column("id", String(), primary_key=True),
            Column("name", String()),
            Column("params", JSON(), nullable=True),
            Column("state", String(), index=True),
            Column("app_name", String(), index=True),
            Column("result", JSON, nullable=True),
            Column("timeout", Integer()),
            Column("result_ttl", Integer()),
            Column(
                "created_at",
                DateTime(),
                index=True,
                server_default=functions.now(),
                nullable=False,
            ),
            Column(
                "updated_at",
                DateTime(),
                server_default=functions.now(),
                nullable=False,
            ),
            extend_existing=True,
        )
        return tbl

    def _create_history_table(self, name):
        tbl = Table(
            name,
            self.meta,
            Column("id", String(), primary_key=True),
            Column("name", String()),
            Column("app_name", String(), index=True),
            Column("state", String()),
            Column(
                "created_at",
                DateTime(),
                index=True,
                server_default=functions.now(),
                nullable=False,
            ),
            extend_existing=True,
        )
        return tbl

    async def add_task(self, task: Task):
        async with self.begin() as conn:
            await conn.execute(self._tasks.insert(), [task.dict()])
            # await conn.commit()

    async def get_task(self, taskid: str) -> Task:
        async with self.conn() as conn:
            stmt = select(self._tasks).where(self._tasks.c.id == taskid).limit(1)
            res = await conn.execute(stmt)
            row = res.fetchone()
            task = Task(**dict(row._mapping))
        return task

    async def list_tasks(self) -> List[Task]:
        async with self.conn() as conn:
            stmt = select(self._tasks)
            res = await conn.execute(stmt)
            rows = res.fetchall()
            tasks = [Task(**dict(r._mapping)) for r in rows]
        return tasks

    async def update_status(self, taskid: str, status: str) -> bool:
        now = datetime.utcnow()

        async with self.begin() as conn:
            stmt = (
                update(self._tasks)
                .where(self._tasks.c.id == taskid)
                .values(state=status, updated_at=now)
            )
            await conn.execute(stmt)

        return True

    async def delete_task(self, taskid: str) -> bool:
        async with self.begin() as conn:
            r = await self._delete(conn, taskid)
        return r

    async def _delete(self, conn, taskid: str) -> bool:
        stmt = sqldelete(self._tasks).where(self._tasks.c.id == taskid)
        await conn.execute(stmt)
        return True

    async def get_result(self, taskid: str) -> Dict[str, Any]:
        async with self.conn() as conn:
            stmt = select(self._tasks).where(self._tasks.c.id == taskid).limit(1)
            res = await conn.execute(stmt)
            row = res.fetchone()
            task_dict = dict(row)
        return task_dict["result"]

    async def set_result(self, taskid: str, *, result: Dict[str, Any], status: str):
        async with self.begin() as conn:
            await self._set_result(conn, taskid, result, status)

    async def _set_result(
        self,
        conn,
        taskid: str,
        result: Dict[str, Any],
        status: str,
    ):
        now = datetime.utcnow()
        stmt = (
            update(self._tasks)
            .where(self._tasks.c.id == taskid)
            .values(result=result, updated_at=now, state=status)
        )
        await conn.execute(stmt)

    async def _vacuum(self):
        await async_vacuum(self.engine, self._tasks.name)

    async def clean_failed(self) -> List[str]:
        async with self.conn() as conn:
            stmt = select(self._tasks).where(
                self._tasks.c.state == TaskStatus.failed.value
            )
            res = await conn.execute(stmt)
            rows = res.fetchall()
            tasks = [Task(**dict(r._mapping)) for r in rows]
        # futures = []
        tasks_ids = []
        async with self.begin() as conn:
            for t in tasks:
                elapsed = elapsed_time_from_finish(t)
                if elapsed > t.result_ttl:
                    # futures.append(self.delete_task(t.id))
                    await self._delete(conn, t.id)
                    tasks_ids.append(t.id)

        # await asyncio.gather(*futures, return_exceptions=True)
        return tasks_ids

    async def _clean_done(self):
        async with self.conn() as conn:
            stmt = select(self._tasks).where(
                self._tasks.c.state == TaskStatus.done.value
            )
            res = await conn.execute(stmt)
            rows = res.fetchall()
            tasks = [Task(**dict(r._mapping)) for r in rows]

        # futures = []
        async with self.begin() as conn:
            for t in tasks:
                elapsed = elapsed_time_from_finish(t)
                if elapsed > t.result_ttl:
                    await self._delete(conn, t.id)
                # futures.append(self.delete_task(t.id))
        # await asyncio.gather(*futures, return_exceptions=False)

    async def _move_to_failed(self):
        async with self.conn() as conn:
            stmt = select(self._tasks).where(
                self._tasks.c.state == TaskStatus.created.value
            )
            res = await conn.execute(stmt)
            rows = res.fetchall()
            tasks = [Task(**dict(r._mapping)) for r in rows]

        # futures = []
        async with self.begin() as conn:
            for t in tasks:
                elapsed = elapsed_time_from_start(t)
                if elapsed > t.timeout:
                    _res = {"error": "timeout, could be running"}
                    # futures.append(
                    #    self.set_result(t.id, result=_res, status=TaskStatus.failed.value)
                    # )
                    await self._set_result(
                        conn, t.id, result=_res, status=TaskStatus.failed.value
                    )
            # await asyncio.gather(*futures, return_exceptions=True)

    async def clean(self):
        await self._clean_done()
        await self._move_to_failed()



from functools import wraps
from inspect import isawaitable
from typing import Dict, List, Optional

from sanic import Request, json

from services import errors
from services.security.base import IAuth


class Authenticator:
    def __init__(
        self,
        app=None,
    ):
        self._scopes = None
        self._require_all = False
        self._validators: Dict[str, IAuth] = {}
        self._validators_list = set()
        if app:
            self.init_app(app)

    def register_validator(self, name: str, validator: IAuth):
        if name not in self._validators_list:
            self._validators.update({name: validator})
            self._validators_list.add(name)

    def list_validators(self) -> List[str]:
        return list(self._validators_list)

    def init_app(self, app):
        app.ctx.auth_beta = self
        app.error_handler.add(errors.web.WebAuthFailed, self.auth_error_handler)
        app.error_handler.add(
            errors.web.MissingAuthorizationHeader, self.auth_error_handler
        )

    def bp_middleware(self, request):
        valid = self.validate(
            self._validators, request, self._scopes, self._require_all
        )
        if not valid:
            raise errors.web.WebAuthFailed()

    def init_bp(
        self,
        bp,
        validators: Dict[str, IAuth],
        scopes: Optional[List[str]] = None,
        require_all=False,
    ):
        self._validators = validators
        self._scopes = scopes
        self._require_all = require_all
        bp.on_request(self.bp_middleware)

    def auth_error_handler(self, request, exception):
        """TODO: move to WebSpec"""
        return json({"msg": "Authentication failed"}, status=401)

    def validate(
        self,
        validators: List[str],
        request,
        policies: List[str] = None,
        require_all=True,
    ) -> bool:
        valid = False
        request.ctx.is_authenticated = False
        request.ctx.is_authorized = False
        request.ctx.user_id = None
        for v in validators:
            auth = self._validators.get(v)
            if auth:
                auth.validate_request(request, policies, require_all)

        if request.ctx.is_authenticated and request.ctx.is_authenticated:
            valid = True
        return valid

    def update_response(self, validators: List[str], request: Request, response):
        for v in validators:
            auth = self._validators.get(v)
            if auth:
                auth.update_response(request, response)


def get_authenticator(request) -> Authenticator:
    return request.app.ctx.auth_beta


def protected(
    validators: List[str], scopes: Optional[List[str]] = None, require_all=True
):
    """verify a token from a request.
    Optionally if a list of scopes is given then will check that scopes
    with the scopes provided by the token.

    :param scopes: a list of scopes
    :param required_all: if true it will check that the all the names provided
    match with the required.

    """

    def decorator(f):
        @wraps(f)
        async def decorated_function(request: Request, *args, **kwargs):
            # auth = get_req_auth(request)
            auth = get_authenticator(request)
            is_valid = auth.validate(validators, request, scopes, require_all)

            if not is_valid:
                raise errors.web.WebAuthFailed(msg="Authentication failed")
            response = f(request, *args, **kwargs)
            if isawaitable(response):
                response = await response

            auth.update_response(validators, request, response)
            return response

        return decorated_function

    return decorator



import json
import os
from datetime import timedelta
from typing import Any, Dict, Generator, List, Union, Optional, AsyncIterator

from google.cloud.storage import Client
from google.cloud.storage.bucket import Bucket as _GoogleBucket
from pydantic import BaseModel
from smart_open import open

from services import defaults, types
from services.storage import IAsyncStore, IStorage
from services.utils import from_async2sync


class BucketConf(BaseModel):
    location: str
    storage_class: str = "STANDARD"
    versioning: bool = False
    labels: Optional[Dict[str, str]] = None


class _GCConf(BaseModel):
    """
    :param CREDENTIALS: credential file open by python
    :param PROJECT: projectid
    :param LOCATION: region/zone
    :param SERVICE_ACCOUNT: client_email
    """

    CREDENTIALS: str
    PROJECT: str
    LOCATION: Optional[str] = None
    SERVICE_ACCOUNT: Optional[str] = None


def get_auth_conf(env_var=defaults.GOOGLE_AUTH_ENV, filepath=None) -> _GCConf:
    """
    https://googleapis.dev/python/google-api-core/latest/auth.html#authentication
    """
    if filepath:
        creds_path = filepath
    else:
        creds_path = os.environ.get(env_var)
    if creds_path:
        with open(creds_path, "r") as f:
            data = json.loads(f.read())
            acc = data["client_email"]
            prj = data["project_id"]
        conf = _GCConf(CREDENTIALS=creds_path, PROJECT=prj, SERVICE_ACCOUNT=acc)
    else:
        conf = _GCConf()
    return conf


class GoogleStore(IStorage):
    """https://googleapis.dev/python/storage/latest/client.html"""

    def __init__(self, bucket: str, opts: Optional[Dict[str, Any]] = None):
        env_var = defaults.GOOGLE_AUTH_ENV
        if opts:
            env_var = opts.get("env_var")

        self._conf = get_auth_conf(env_var=env_var)
        self._bucket = bucket
        self.client = Client.from_service_account_json(self._conf.CREDENTIALS)
        self.bucket: _GoogleBucket = None
        self._params = {"client": self.client}

    @property
    def uri(self):
        return f"gs://{self._bucket}"

    def _current_or_get_bucket(self) -> _GoogleBucket:
        if not self.bucket:
            self.bucket = self.client.get_bucket(self._bucket)
        return self.bucket

    def _to_bucket(self, b: _GoogleBucket) -> types.Bucket:
        _bucket = types.Bucket(
            name=b.name,
            storage_class=b.storage_class,
            location=b.location,
            url=f"gs://{b.name}",
            versioning=b.versioning_enabled,
            labels=b.labels,
            created_at=b.time_created,
        )
        return _bucket

    def get_bucket(self) -> types.Bucket:
        b = self._current_or_get_bucket()
        return self._to_bucket(b)

    def create_bucket(self, opts: Optional[Dict[str, str]] = None) -> types.Bucket:
        conf = BucketConf(**opts)

        _b = self.client.create_bucket(
            name=self._bucket,
            storage_class=conf.storage_class,
            location=conf.location,
            labels=conf.labels,
        )
        _bucket = self._to_bucket(_b)
        self.bucket = _b
        return _bucket

    def put(self, key: str, bdata: bytes):
        # obj = io.BytesIO(bdata)
        blob = self.bucket.blob(key)
        blob.upload_from_string(bdata, content_type="application/octet-stream")

    def _writer(self, key: str, generator: Generator[bytes, None, None]):
        with open(f"{self.uri}/{key}", "wb", transport_params=self._params) as f:
            data = True
            while data:
                try:
                    _data = next(generator)
                    f.write(_data)
                except StopIteration:
                    data = False

    def list(self) -> List[str]:
        # TODO: define a type for objects that include size, name, path etc...
        blobs = [b.name for b in self.bucket.list_blobs()]
        return blobs

    def delete(self, key: str):
        self.bucket.delete_blob(key)

    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        rsp = True
        try:
            self._writer(key, generator)
        except Exception:
            rsp = False
        return rsp

    def get(self, key: str) -> Union[bytes, None]:
        blob = self.bucket.blob(key)
        obj = None
        try:
            obj = blob.download_as_bytes()
        except Exception:
            pass
        return obj

    def get_stream(self, key: str) -> Generator[bytes, None, None]:
        uri = f"{self.uri}/{key}"
        for chunk in open(uri, "rb", transport_params=self._params):
            yield chunk

    def download_signed(self, key, minutes=15, bucket=None) -> str:
        _b = self._current_or_get_bucket()
        blob = _b.blob(key)
        url = blob.generate_signed_url(
            version="v4",
            expiration=timedelta(minutes=minutes),
            method="GET",
        )
        return url

    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        """
        curl -X PUT -H 'Content-Type: application/octet-stream' "
        "--upload-file my-file '{url}'
        """

        _b = self._current_or_get_bucket()
        blob = _b.blob(key)
        url = blob.generate_signed_url(
            version="v4",
            expiration=timedelta(minutes=minutes),
            method="PUT",
            content_type=content_type,
        )
        return url


class AsyncGoogleStore(IAsyncStore):
    """A hacky solution because thereisn't trustworthy async lib"""

    def __init__(self, bucket: str, client: GoogleStore):
        self._bucket = bucket
        self.client = client
        self.bucket: _GoogleBucket = client.bucket

    @classmethod
    async def from_uri(
        cls, bucket, opts: Optional[Dict[str, Any]] = None
    ) -> "IAsyncStore":
        c = GoogleStore(bucket, opts)
        await from_async2sync(c.get_bucket)
        bucket = c.bucket
        return cls(bucket=bucket, client=c)

    async def get_bucket(self) -> types.Bucket:
        return await from_async2sync(self.client.get_bucket)

    async def create_bucket(
        self, opts: Optional[Dict[str, str]] = None
    ) -> types.Bucket:
        return await from_async2sync(self.client.create_bucket, opts)

    async def _current_or_get_bucket(self) -> _GoogleBucket:
        if not self.bucket:
            self.bucket = await from_async2sync(self.client.get_bucket)
        return self.bucket

    async def put(self, key: str, bdata: bytes):
        await from_async2sync(self.client.put, key, bdata)

    async def put_stream(
        self, key: str, generator: Generator[bytes, None, None]
    ) -> bool:
        rsp = await from_async2sync(self.client.put_stream, key, generator)
        return rsp

    async def get(self, key: str) -> Union[bytes, str, None]:
        rsp = await from_async2sync(self.client.get, key)
        return rsp

    async def get_stream(self, key: str, buffer_size=1024) -> AsyncIterator:
        yield await from_async2sync(self.client.get_stream, key)

    async def list(self) -> List[str]:
        rsp = await from_async2sync(self.client.list)
        return rsp

    async def delete(self, key: str):
        await from_async2sync(self.client.delete, key)

    def download_signed(self, key, minutes=15, bucket=None) -> str:
        _b = self._current_or_get_bucket()
        blob = _b.blob(key)
        url = blob.generate_signed_url(
            version="v4",
            expiration=timedelta(minutes=minutes),
            method="GET",
        )
        return url

    def upload_signed(
        self, key, minutes=15, bucket=None, content_type="application/octed-stream"
    ) -> str:
        """
        curl -X PUT -H 'Content-Type: application/octet-stream' "
        "--upload-file my-file '{url}'
        """

        _b = self._current_or_get_bucket()
        blob = _b.blob(key)
        url = blob.generate_signed_url(
            version="v4",
            expiration=timedelta(minutes=minutes),
            method="PUT",
            content_type=content_type,
        )
        return url



from typing import Dict, Optional
from urllib.parse import urljoin

from jinja2 import Environment
from jinja2.utils import markupsafe

from .simple_tags import StandaloneTag

scripts_attrs = {"type": "module", "async": "", "defer": ""}


def generate_script_tag(
    src: str, attrs: Optional[Dict[str, str]] = None
) -> str:
    """Generates an HTML script tag."""
    attrs_str: str = ""
    if attrs is not None:
        for key, value in attrs.items():
            if value:
                attrs_str += f"{key}={value} "
            else:
                attrs_str += f"{key} "

    return f'<script {attrs_str.strip()} src="{src}"></script>'


def generate_stylesheet_tag(href: str) -> str:
    """
    Generates and HTML <link> stylesheet tag for CSS.
    Arguments:
        href {str} -- CSS file URL.
    Returns:
        str -- CSS link tag.
    """
    return '\t<link rel="stylesheet" href="{href}" />'.format(href=href)


def generate_react_hmr():
    return f"""
        <script type="module">
        import RefreshRuntime from '{cls.generate_vite_server_url()}@react-refresh'
        RefreshRuntime.injectIntoGlobalHook(window)
        window.$RefreshReg$ = () => {{}}
        window.$RefreshSig$ = () => (type) => type
        window.__vite_plugin_react_preamble_installed__=true
        </script>
                """


class ViteDev(StandaloneTag):
    tags = {"vite_dev"}

    def __init__(self, environment):
        super().__init__(environment)
        # environment.extend(vite_metadata_prefix="", vite_metadata=None)
        environment.extend(vite_dev_mod_prefix="", vite_dev_mode=None)
        environment.extend(vite_dev_server_prefix="", vite_dev_server=None)

    @staticmethod
    def generate_vite_server_url(vite_dev_srv: str, path: Optional[str] = None):
        return urljoin(
            vite_dev_srv, path if path is not None else "")

    def render(self, script_name="main.js"):
        _url = f"{self.environment.vite_dev_server}/{script_name}"
        tags = []
        if self.environment.vite_dev_mode:
            tag = generate_script_tag(_url, {"type": "module"})
            tags.append(tag)
        if self.environment.vite_react_mode:
            tag = generate_react_hmr()
            tags.append(tag)
        return markupsafe.Markup("\n".join(tags))


class ViteAsset(StandaloneTag):
    tags = {"vite_asset"}

    def __init__(self, environment):
        super().__init__(environment)
        # environment.extend(vite_metadata_prefix="", vite_metadata=None)
        environment.extend(vite_manifest_prefix="", vite_manifest=None)
        environment.extend(vite_dev_mod_prefix="", vite_dev_mode=None)

    def render(self, asset_name="main.js"):
        if not self.environment.vite_dev_mode:
            m = self.environment.vite_manifest[asset_name]
            tags = []
            asset = m["file"]
            tag = generate_script_tag(asset, attrs=scripts_attrs)
            tags.append(tag)
            if m.get("css"):
                for css in m.get("css"):
                    tags.append(
                        generate_stylesheet_tag(css)
                    )

            return markupsafe.Markup("\n".join(tags))
        return markupsafe.Markup()



from services.db import AsyncSQL, SQL
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase, mapped_column, Mapped
from sqlalchemy.ext.compiler import compiles
from services.db import SQL, AsyncSQL
from services.db.mixins import BaseMixinAsDataclass
from sqlalchemy.types import BINARY, JSON, BigInteger

# equivalent to Base = declarative_base()


class Base(DeclarativeBase):
    pass


class TestModel(BaseMixinAsDataclass, Base):
    __test__ = False
    __tablename__ = "testing"
    __mapper_args__ = {"eager_defaults": True}
    fullname: Mapped[str] = mapped_column(String)


async def async_create_all(db: AsyncSQL):
    await db.create_all(Base.metadata)


def create_all(db: SQL):
    db.create_all(Base.metadata)


@compiles(JSON, "postgresql")
def compile_jsonb_postgres(type_, compile, **kw):
    return "JSONB"


@compiles(BINARY, "postgresql")
def compile_bytea_postgres(type_, compiler, **kw):
    return "BYTEA"


@compiles(BigInteger, "sqlite")
def compile_integer_sqlite(type_, compiler, **kw):
    return "Integer"



import sys
from services import defaults, types

# WARNING:
# We do our best effort to keep sensible information private
# but in the scenario of an intrusion into the network or machines
# where servers are hosted, then it represents a risk for the information
# stored in those machines.

# General Folders for the server
DATABASES = {
    "default": types.Database(
        name="default",
        async_url="sqlite+aiosqlite:///db.sqlite",
        sync_url="sqlite:///db.sqlite",
        description="Default database",
    )
}

APPS = [
    "example.web.WebApp",
]

SECURITY = types.SecurityConfig(
    secret_key="ZXigJdqSQXwPqy7XTCAb1ztgM0_dynZT8i21g2JtjLk",
    jwt=types.JWTConfig(
        alg="HS256",
        secret="ZXigJdqSQXwPqy7XTCAb1ztgM0_dynZT8i21g2JtjLk"
    )
)

TASKS = types.TasksBackend(
    uri="sqlite+aiosqlite:///tasks.sqlite",
    backend_class="services.ext.sql.workers.SQLBackend"

)
STORAGE = {
    "default": types.Storage(
        bucket=".storage", store_class="services.storage.AsyncLocal"
    )
}
TEMPLATES_DIR = [
    "example/templates"
]



import contextlib
from typing import AsyncIterator
from sanic import Sanic
from sanic.log import logger
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio.engine import AsyncEngine

from services.db.helpers import AsyncSQL
from services.types import Settings

# from typing import AsyncContextManager


class DBHelper:
    def __init__(self, app: Sanic):
        self.app = app

    async def listener_db(self, app: Sanic):
        for k, v in app.config.DATABASES.items():
            _db = AsyncSQL.from_conf(v)
            logger.info("Starting db %s", k)
            # await _db.init()
            app.ctx.databases[k] = _db

    def get_db(self, *, name="default") -> AsyncSQL:
        return self.app.ctx.databases.get(name)

    def get_engine(self, *, name="default") -> AsyncEngine:
        db = self.get_db(name=name)
        return db.engine

    async def dispose(self, *, name="default"):
        db = self.get_db(name=name)
        await db.dispose()

    @contextlib.asynccontextmanager
    async def session(self, *, name="default") -> AsyncIterator[AsyncSession]:
        """
        To be used for ORM implementation
        """
        db = self.get_db(name=name)
        try:
            async with db.async_session() as session:
                yield session
        finally:
            await self.dispose(name=name)

    @contextlib.asynccontextmanager
    async def conn(self, *, name="default") -> AsyncIterator[AsyncSession]:
        """
        To be used as the core part
        """
        db = self.get_db(name=name)
        try:
            async with db.conn() as conn:
                yield conn
        finally:
            await self.dispose(name=name)

    @contextlib.asynccontextmanager
    async def begin(self, *, name="default") -> AsyncIterator[AsyncSession]:
        """
        To be used as the core part
        """
        db = self.get_db(name=name)
        try:
            async with db.begin() as conn:
                yield conn
        finally:
            await self.dispose(name=name)


def init_db(app: Sanic, settings: Settings):
    app.config.DATABASES = settings.DATABASES
    db_helper = DBHelper(app)
    app.register_listener(db_helper.listener_db, "before_server_start")
    app.ext.dependency(db_helper)



from datetime import datetime
from typing import Any, Dict, Generic, List, TypeVar, Union

from cryptography.fernet import Fernet
from sqlalchemy import delete as sqldelete
from sqlalchemy import desc, func, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.attributes import InstrumentedAttribute

from services import types
from services.errors import DBObjectNotFound
from services.utils import get_class

ModelT = TypeVar("ModelT")


class AsyncManagerBase(Generic[ModelT]):
    model_class: str
    lookup_key: str
    active_key: str = "is_active"

    def __init__(self):
        self._model: ModelT = get_class(self.model_class)
        self._key = self.lookup_key
        self._active_key = self.active_key

    @property
    def tablename(self) -> str:
        return self._model.__tablename__

    def _get_column(self, col_name) -> InstrumentedAttribute:
        return getattr(self._model, col_name)

    async def list(
        self, session, order_by=None, is_active=True, offset=0, limit=10
    ) -> List[ModelT]:
        stmt = select(self._model).offset(offset).limit(limit)
        if is_active:
            attr = self._get_column(self._active_key)
            stmt = stmt.where(attr == True)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = await session.execute(stmt)
        return list(result.scalars())

    def _obj_or_raise(self, lookup, obj: Union[ModelT, None], is_active=True) -> ModelT:
        if not obj:
            raise DBObjectNotFound(self.tablename, lookup)
        if is_active:
            attr = self._get_column(self._active_key)
            if not attr:
                raise DBObjectNotFound(self.tablename, lookup)

        return obj

    async def _get_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = await session.execute(stmt)

        return rsp.scalar_one_or_none()

    async def _commit_or_rollback(self, session) -> bool:
        try:
            await session.commit()
        except IntegrityError:
            await session.rollback()
            return False
        return True

    async def count(self, session, is_active=True) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        if is_active:
            attr = self._get_column(self.active_key)
            stmt = stmt.where(attr == True)
        res = await session.execute(stmt)
        return res.scalar()

    async def create(
        self, session, data: Dict[str, Any], commit=True
    ) -> Union[ModelT, None]:
        model = self._model(**data)
        session.add(model)
        if commit:
            res = await self._commit_or_rollback(session)
            if not res:
                return None
        return model

    async def get(self, session, key: str, is_active=True) -> ModelT:
        _obj = await self._get_one(session, key)
        return self._obj_or_raise(key, _obj, is_active)

    async def update(self, session, key: str, data: Dict[str, Any]):
        u = await self._get_one(session, key)
        if not u:
            raise DBObjectNotFound(self.tablename, key)

        for k, v in data.items():
            setattr(u, k, v)
        u.updated_at = datetime.utcnow()
        session.add(u)

    async def _delete_hard(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        await session.execute(stmt)

    async def _delete_flag(self, session, key: str):
        obj = await self._get_one(session, key)
        setattr(obj, self._active_key, False)
        obj.updated_at = datetime.utcnow()
        session.add(obj)

    async def delete(self, session, key: str, hard=False):
        if hard:
            await self._delete_hard(session, key)
        else:
            await self._delete_flag(session, key)


class ManagerBase(Generic[ModelT]):
    model_class: str
    lookup_key: str
    active_key: str = "is_active"

    def __init__(self):
        self._model: ModelT = get_class(self.model_class)
        self._key = self.lookup_key
        self._active_key = self.active_key

    @property
    def tablename(self) -> str:
        return self._model.__tablename__

    def _get_column(self, col_name) -> InstrumentedAttribute:
        return getattr(self._model, col_name)

    async def list(
        self, session, order_by=None, is_active=True, offset=0, limit=10
    ) -> List[ModelT]:
        stmt = select(self._model).offset(offset).limit(limit)
        if is_active:
            attr = self._get_column(self._active_key)
            stmt = stmt.where(attr == True)
        if order_by:
            stmt.order_by(order_by, desc(order_by))
        result = await session.execute(stmt)
        return list(result.scalars())

    def _obj_or_raise(self, lookup, obj: Union[ModelT, None], is_active=True) -> ModelT:
        if not obj:
            raise DBObjectNotFound(self.tablename, lookup)
        if is_active:
            attr = self._get_column(self._active_key)
            if not attr:
                raise DBObjectNotFound(self.tablename, lookup)

        return obj

    def _get_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = session.execute(stmt)

        return rsp.scalar_one_or_none()

    def _commit_or_rollback(self, session) -> bool:
        try:
            session.commit()
        except IntegrityError:
            session.rollback()
            return False
        return True

    def count(self, session, is_active=True) -> int:
        stmt = select(func.count(self._model.__table__.c.id))
        if is_active:
            attr = self._get_column(self.active_key)
            stmt = stmt.where(attr == True)
        res = session.execute(stmt)
        return res.scalar()

    def create(self, session, data: Dict[str, Any], commit=True) -> Union[ModelT, None]:
        model = self._model(**data)
        session.add(model)
        if commit:
            res = self._commit_or_rollback(session)
            if not res:
                return None
        return model

    def get(self, session, key: str, is_active=True) -> ModelT:
        _obj = self._get_one(session, key)
        return self._obj_or_raise(key, _obj, is_active)

    def update(self, session, key: str, data: Dict[str, Any]):
        u = self._get_one(session, key)
        if not u:
            raise DBObjectNotFound(self.tablename, key)

        for k, v in data.items():
            setattr(u, k, v)
        u.updated_at = datetime.utcnow()
        session.add(u)

    def _delete_hard(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        session.execute(stmt)

    def _delete_flag(self, session, key: str):
        obj = self._get_one(session, key)
        setattr(obj, self._active_key, False)
        obj.updated_at = datetime.utcnow()
        session.add(obj)

    def delete(self, session, key: str, hard=False):
        if hard:
            self._delete_hard(session, key)
        else:
            self._delete_flag(session, key)


class AsyncSecretBase(Generic[ModelT]):
    model_class: str
    lookup_key: str

    def __init__(self, private_key: str):
        self._private_key = private_key
        self._f = Fernet(self._private_key)
        self._model: ModelT = get_class(self.model_class)
        self._key = self.lookup_key

    def decrypt(self, encrypted: bytes) -> str:
        e = self._f.decrypt(encrypted).decode("utf-8")
        return e

    def encrypt(self, text: str) -> bytes:
        _hash = self._f.encrypt(text.encode("utf-8"))
        return _hash

    def _get_column(self, col_name) -> InstrumentedAttribute:
        return getattr(self._model, col_name)

    async def _get_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = await session.execute(stmt)

        return rsp.scalar_one_or_none()

    @property
    def tablename(self) -> str:
        return self._model.__tablename__

    async def delete(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        await session.execute(stmt)

    def _obj_or_raise(self, lookup, obj: Union[ModelT, None]) -> ModelT:
        if not obj:
            raise DBObjectNotFound(self.tablename, lookup)
        return obj

    async def get(self, session, key: str) -> str:
        _obj = await self._get_one(session, key)
        encrypted = self._obj_or_raise(key, _obj)
        return self.decrypt(encrypted)

    def write(self, session, key: str, data: str) -> ModelT:
        enc = self.encrypt(data)
        model = self._model(key=key, secret=enc)
        session.add(model)
        return model


class SecretBase(Generic[ModelT]):
    model_class: str
    lookup_key: str

    def __init__(self, settings: types.Settings):
        self._private_key = settings.SECURITY.secret_key
        self._f = Fernet(self._private_key)
        self._model: ModelT = get_class(self.model_class)
        self._key = self.lookup_key

    def decrypt(self, encrypted: bytes) -> str:
        e = self._f.decrypt(encrypted).decode("utf-8")
        return e

    def encrypt(self, text: str) -> bytes:
        _hash = self._f.encrypt(text.encode("utf-8"))
        return _hash

    def _get_column(self, col_name) -> InstrumentedAttribute:
        return getattr(self._model, col_name)

    def _get_one(self, session, key: str) -> ModelT:
        attr_key = self._get_column(self._key)
        stmt = select(self._model).where(attr_key == key).limit(1)
        rsp = session.execute(stmt)

        return rsp.scalar_one_or_none()

    @property
    def tablename(self) -> str:
        return self._model.__tablename__

    def delete(self, session, key: str):
        attr_key = self._get_column(self._key)
        stmt = sqldelete(self._model).where(attr_key == key)
        session.execute(stmt)

    def _obj_or_raise(self, lookup, obj: Union[ModelT, None]) -> ModelT:
        if not obj:
            raise DBObjectNotFound(self.tablename, lookup)
        return obj

    def get(self, session, key: str) -> str:
        _obj = self._get_one(session, key)
        encrypted = self._obj_or_raise(key, _obj)
        return self.decrypt(encrypted)

    def write(self, session, key: str, data: str) -> ModelT:
        enc = self.encrypt(data)
        model = self._model(key=key, secret=enc)
        session.add(model)
        return model



import time
from pydantic import BaseModel


class TaskExample(BaseModel):
    do: str
    wait: int


def dummy(t: TaskExample):
    print(f"=> The task will do: {t.do}, and it will wait for {t.wait}")
    time.sleep(t.wait)
    print(f"=> Dummy's task did")


from typing import Optional, Union

import redis.asyncio as redis

from services import types
from services.security.base import ITokenStore


class RedisTokenStore(ITokenStore[redis.client.Redis]):
    ns = "rtk"

    @classmethod
    async def from_conf(
        cls, conf: types.SecurityConfig, driver: Optional[redis.client.Redis] = None
    ) -> "RedisTokenStore":
        if not driver:
            driver = await redis.from_url(conf.token_store_uri)

        return cls(conf, driver)

    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        await self.driver.set(f"{self.ns}:{key}", value)
        return True

    async def get(self, key: str) -> str:
        val = await self.driver.get(f"{self.ns}:{key}")
        value = val
        if not isinstance(val, str) and val:
            value = val.decode("utf-8")

        return value

    async def delete(self, key: str):
        await self.driver.delete(f"{self.ns}:{key}")



import pytest
from sqlalchemy.engine.result import ChunkedIteratorResult

from services.db import GenericManager

from .common import TestModel


class Manager(GenericManager[TestModel]):
    model_class = "tests.common.TestModel"
    lookup_key = "id"


def test_managers2_generic(db):
    mg = Manager()
    assert isinstance(mg, GenericManager)


def test_managers2_create(db):
    mg = Manager()
    with db.session() as s:
        obj = mg.create(s, {"fullname": "hello"}, commit=True)
        row = mg.get_one(s, obj.id)

    assert isinstance(mg, GenericManager)
    assert isinstance(obj, TestModel)
    assert isinstance(row, TestModel)
    # assert isinstance(row, TestModel)


def test_managers2_list(db):
    mg = Manager()
    with db.session() as s:
        all_ = mg.list(s)
        totals = mg.scalars(all_)

    assert isinstance(all_, ChunkedIteratorResult)
    assert len(totals) > 0
    assert isinstance(totals[0], TestModel)


@pytest.mark.asyncio
async def test_managers2_acreate(adb):
    mg = Manager()
    async with adb.session() as s:
        obj = await mg.acreate(s, {"fullname": "pepe"}, commit=True)
        row = await mg.aget_one(s, obj.id)

    assert isinstance(mg, GenericManager)
    assert isinstance(obj, TestModel)
    assert isinstance(row, TestModel)


@pytest.mark.asyncio
async def test_managers2_alist(adb):
    mg = Manager()
    async with adb.session() as s:
        await mg.acreate(s, {"fullname": "pepe"}, commit=True)
        all_ = await mg.alist(s)
        totals = mg.scalars(all_)

    assert isinstance(all_, ChunkedIteratorResult)
    assert len(totals) > 0
    assert isinstance(totals[0], TestModel)



from sanic import Sanic

from services.base import ViewSet, WebAppSpec
from services.security import Authenticator
from services.security.jwtauth import JWTAuth
from services.security.sessionauth import SessionAuth
from services.security.memory_store import MemoryTokenStore
from services.types import Settings
{% if data.tasks -%}
from services import workers
{% endif -%}

{% if data.html -%}
web = ViewSet(
    blueprints=["views"],
    package="{{ data.app_name }}"
)
{% endif -%}
{% if data.users %}
# from services.users import UserManager
from .managers import UserManager, GroupManager

api = ViewSet(
        blueprints=["users", "{{ data.app_name }}"],
        package="{{ data.app_name }}.api"
)


class WebApp(WebAppSpec):
    name = "{{ data.app_name }}"
    {% if data.html -%}
    views = [api, web]
    {% else %}
    views = [api]
    {% endif -%}
    def hook_users(self, app: Sanic):
        """
        delete if you don't want to use the pre-built user system
        """
        settings = self.get_app_settings(app)
        gm = GroupManager()
        um = UserManager(salt=settings.SECURITY.secret_key, groups=gm)
        app.ext.dependency(um)
        db = self.get_db(app, settings.USER_DB)

    def init(self, app: Sanic, settings: Settings):
        """ complete with your own logic """
        app.register_listener(self.hook_users, "before_server_start")
        store = MemoryTokenStore(settings.SECURITY)
        jwtauth = JWTAuth(settings.SECURITY, store)
        session_auth = SessionAuth(settings.SECURITY, secure=False)
        app.config.JWT_ALLOW_REFRESH = settings.SECURITY.jwt.allow_refresh_token
        self.register_auth_validator(app, "jwt", jwtauth)
        self.register_auth_validator(app, "cookie", session_auth)
        {% if data.tasks -%}
        if not settings.SINGLE_PROCESS:
            _q_conf = workers.QueueConfig(
                app_name=self.name, qname="default", backend=settings.TASKS)
            workers.create(app, _q_conf)
            workers.TaskQueue.setup(app, _q_conf)
        {% endif -%}

        # worker = Dummy(proc_name="DummyWorker")
        # worker.init_app(app)

        self.init_blueprints(app)
{% else %}

api = ViewSet(
        blueprints=["{{ data.app_name }}"],
        package="{{ data.app_name }}.api"
)

class WebApp(WebAppSpec):
    name = "{{ data.app_name }}"
    {% if data.html -%}
    views = [api, web]
    {% else %}
    views = [api]
    {% endif -%}

    def init(self, app: Sanic, settings: Settings):
        """ complete with your own logic """

        store = MemoryTokenStore(settings.SECURITY)
        jwtauth = JWTAuth(settings.SECURITY, store)
        app.config.JWT_ALLOW_REFRESH = settings.SECURITY.jwt.allow_refresh_token
        self.register_auth_validator(app, "jwt", jwtauth)
        self.init_blueprints(app)
        {% if data.tasks -%}
        if not settings.SINGLE_PROCESS:
            _q_conf = workers.QueueConfig(
                app_name=self.name, qname="default", backend=settings.TASKS)
            workers.create(app, _q_conf)
            workers.TaskQueue.setup(app, _q_conf)
        {% endif -%}

        

{% endif %}



import sys
from pathlib import Path
from secrets import token_urlsafe

import click
from rich.console import Console

from services import init_script

console = Console()


@click.command(name="create-app")
@click.option(
    "--html/--no-html",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add templates and html view example",
)
@click.option(
    "--vite/--no-vite",
    default=False,
    is_flag=True,
    show_default=True,
    help="Get support for Vite. Html feature should be enabled",
)
@click.option(
    "--users/--no-users",
    default=False,
    is_flag=True,
    show_default=True,
    help="Provide the user system",
)
@click.option(
    "--tasks/--no-tasks",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add a dummy task and the worker as example",
)
@click.option(
    "--sql/--no-sql",
    default=False,
    is_flag=True,
    show_default=True,
    help="Add DB related files",
)
@click.argument("appname")
def create_app_cli(appname, vite, tasks, users, sql, html):
    """Create the structure of an app"""
    root = Path.cwd()
    opts = init_script.ScriptOpts(
        base_path=root,
        secret_key=token_urlsafe(32),
        app_name=appname,
        users=users,
        vite_enabled=vite,
        tasks=tasks,
        sql=sql,
        html=html,
    )
    init_script.create_app(opts)
    console.print(f"[green bold]App {appname} created.[/]")



from typing import Any, Dict
from pydantic import BaseModel
from sqlalchemy import BigInteger, Column, String

# from services.db.psql_search import WithSearchField, tsvector_column
from example.users_models import UserModel, GroupModel, user_group_table
from .db import Base


# class SearchExampleModel(Base, WithSearchField):
#     __tablename__ = "example_example"
#     __mapper_args__ = {"eager_defaults": True}
#
#     id = Column(BigInteger, primary_key=True)
#     name = Column(String())
#     title = Column(String())
#     description = Column(String())
#     tsvector = tsvector_column()


class ExampleModel(Base):
    __tablename__ = "example_example"
    __mapper_args__ = {"eager_defaults": True}

    id = Column(BigInteger, primary_key=True)
    name = Column(String())
    title = Column(String())
    description = Column(String())
